# Implements a minimal reader for OGIP PHA format for spectral data
# (https://heasarc.gsfc.nasa.gov/docs/heasarc/ofwg/docs/spectra/ogip_92_007/node6.html)

# Author: Giacomo Vianello (giacomov@slac.stanford.edu)

import os
import warnings
from math import log

import matplotlib.pyplot as plt

from threeML.io.step_plot import step_plot
from threeML.plugins.gammaln import logfactorial


import astropy.io.fits as pyfits
import numpy as np
from astromodels.parameter import Parameter

from threeML.io import file_utils

requiredKeywords = {}
requiredKeywords['observed'] = ("mission:TELESCOP,instrument:INSTRUME,filter:FILTER," +
                                "exposure:EXPOSURE,backfile:BACKFILE," +
                                "corrfile:CORRFILE,corrscal:CORRSCAL,respfile:RESPFILE," +
                                "ancrfile:ANCRFILE,hduclass:HDUCLASS," +
                                "hduclas1:HDUCLAS1,hduvers:HDUVERS,poisserr:POISSERR," +
                                "chantype:CHANTYPE,n_channels:DETCHANS").split(",")
requiredKeywords['background'] = ("mission:TELESCOP,instrument:INSTRUME,filter:FILTER," +
                                  "exposure:EXPOSURE," +
                                  "hduclass:HDUCLASS," +
                                  "hduclas1:HDUCLAS1,hduvers:HDUVERS,poisserr:POISSERR," +
                                  "chantype:CHANTYPE,n_channels:DETCHANS").split(",")
mightBeColumns = {}
mightBeColumns['observed'] = ("EXPOSURE,BACKFILE," +
                              "CORRFILE,CORRSCAL," +
                              "RESPFILE,ANCRFILE").split(",")
mightBeColumns['background'] = ("EXPOSURE").split(",")


class OGIPPluginBase(object):

    def __init__(self, name, response, **other_datafiles):

        self.name = name

        # Check that all file exists

        all_files = list(other_datafiles.values())
        all_files.append(response)

        for filename in all_files:

            if not file_utils.file_existing_and_readable(filename.split("{")[0]):
                raise IOError("File %s does not exists!" % filename)

        self.response = Response(response, other_datafiles.get("arffile"))

        self.counts = None
        self.bkgCounts = None
        self.mask = None
        self.exposure = None

    def _initialSetup(self, mask, counts, bkgCounts, exposure, bkgErr):

        self.mask = mask
        self.counts = counts
        self.bkgCounts = bkgCounts
        self.exposure = exposure
        self.bkgErr = bkgErr

        # Check that counts is positive
        idx = (self.counts < 0)

        if (np.sum(idx) > 0):
            warnings.warn("The observed spectrum for %s " % self.name +
                          "has negative channels! Fixing those to zero.",
                          RuntimeWarning)
            self.counts[idx] = 0

        pass

        # Get the background counts for this spectrum
        # self.bkgCounts = (self.bkgfile.getRates()[self.mask]
        #                  * self.exposure)

        # Check that bkgCounts is positive

        idx = (self.bkgCounts < 0)

        if (np.sum(idx) > 0):
            warnings.warn("The background spectrum for %s " % self.name +
                          "has negative channels! Fixing those to zero.",
                          RuntimeWarning)
            self.bkgCounts[idx] = 0

        # Check that errors are zeros when the bkg is zero
        if (bkgErr is not None) and np.any((self.bkgErr == 0) & (self.bkgCounts > 0)):
            raise RuntimeError("The background error is zero but the background counts are not zero for some bins. "
                               "Data are corrupted.")

        # Check that the observed counts are positive

        idx = self.counts < 0

        if np.sum(idx) > 0:
            raise RuntimeError("Negative counts in observed spectrum %s. Data are corrupted." % ('fix this'))

        # Keep a copy which will never be modified
        self.counts_backup = np.array(self.counts, copy=True)
        self.bkgCounts_backup = np.array(self.bkgCounts, copy=True)

        if bkgErr is not None:
            self.bkgErr_backup = np.array(self.bkgErr, copy=True)

        self.counts = self.counts[self.mask]
        self.bkgCounts = self.bkgCounts[self.mask]

        if bkgErr is not None:

            self.bkgErr = self.bkgErr[self.mask]

            # Check that errors are zeros when the bkg is zero
            if np.any((self.bkgErr == 0) & (self.bkgCounts > 0)):
                raise RuntimeError("The background error is zero but the background counts are not zero for some bins. "
                                   "Data are corrupted.")

        # Effective area correction is disabled by default, i.e.,
        # the nuisance parameter is fixed to 1
        self.nuisanceParameters = {}
        self.nuisanceParameters['InterCalib'] = Parameter("InterCalib", 1, min_value=0.9, max_value=1.1, delta=0.01)
        self.nuisanceParameters['InterCalib'].fix = True

    def set_model(self, likelihoodModel):
        '''
        Set the model to be used in the joint minimization.
        '''
        self.likelihoodModel = likelihoodModel

        diffFlux, integral = self._get_diff_flux_and_integral()

        # This is a wrapper which iterates over all the point sources and get
        # the fluxes
        # We assume there are no extended sources, since we cannot handle them here

        assert self.likelihoodModel.get_number_of_extended_sources() == 0, "OGIP-like plugins do not support " \
                                                                           "extended sources"



        self.response.set_function(diffFlux, integral)

    def _get_diff_flux_and_integral(self):

        nPointSources = self.likelihoodModel.get_number_of_point_sources()

        def diffFlux(energies):
            fluxes = self.likelihoodModel.get_point_source_fluxes(0, energies)

            # If we have only one point source, this will never be executed
            for i in range(1, nPointSources):
                fluxes += self.likelihoodModel.get_point_source_fluxes(i, energies)

            return fluxes

        # The following integrates the diffFlux function using Simpson's rule
        # This assume that the intervals e1,e2 are all small, which is guaranteed
        # for any reasonable response matrix, given that e1 and e2 are Monte-Carlo
        # energies. It also assumes that the function is smooth in the interval
        # e1 - e2 and twice-differentiable, again reasonable on small intervals for
        # decent models. It might fail for models with too sharp features, smaller
        # than the size of the monte carlo interval.

        def integral(e1, e2):

            # Simpson's rule

            return (e2 - e1) / 6.0 * (diffFlux(e1)
                                      + 4 * diffFlux((e1 + e2) / 2.0)
                                      + diffFlux(e2))

        return diffFlux, integral

    def get_name(self):
        '''
        Return a name for this dataset (likely set during the constructor)
        '''
        return self.name

    def inner_fit(self):

        # ad the moment there is no fitting here. We might introduce the renormalization constant
        # at some point

        return self.get_log_like()

    def get_folded_model(self):

        # Get the folded model for this spectrum
        # (this is the rate predicted, in cts/s)

        return self.response.convolve()[self.mask]

    def _get_model_counts(self):

        # Get the folded model for this spectrum (this is the rate predicted,
        # in cts/s)

        folded = self.get_folded_model()

        # Model is folded+background (i.e., we assume negligible errors on the
        # background)
        modelCounts = self.nuisanceParameters['InterCalib'].value * folded * self.exposure

        # Put all negative predicted counts to zero
        clippedModelCounts = np.clip(modelCounts, 0, 1e10)

        return clippedModelCounts

    def set_active_measurements(self, *args):
        '''Set the measurements to be used during the analysis.
        Use as many ranges as you need,
        specified as 'emin-emax'. Energies are in keV. Example:

        setActiveMeasurements('10-12.5','56.0-100.0')

        which will set the energy range 10-12.5 keV and 56-100 keV to be
        used in the analysis'''

        # To implelemnt this we will use an array of boolean index,
        # which will filter
        # out the non-used channels during the logLike

        # Now build the new mask: values for which the mask is 0 will be masked
        mask = np.zeros_like(self.counts_backup)

        for arg in args:

            ee = map(float, arg.replace(" ", "").split("-"))
            emin, emax = sorted(ee)

            idx1 = self.response.energy_to_channel(emin)
            idx2 = self.response.energy_to_channel(emax)

            mask[idx1:idx2 + 1] = True

        self.mask = np.array(mask, np.bool)

        self.counts = self.counts_backup[self.mask]
        self.bkgCounts = self.bkgCounts_backup[self.mask]

        if self.bkgErr is not None:

            self.bkgErr = self.bkgErr_backup[self.mask]

        print("Now using %s channels out of %s" % (np.sum(self.mask),
                                                   self.mask.shape[0]
                                                   ))

    def _compute_log_like(self, modelCounts):

        raise NotImplementedError("You have to override this with the proper statistic")

    def get_log_like(self):
        '''
        Return the value of the log-likelihood with the current values for the
        parameters
        '''

        modelCounts = self._get_model_counts()

        logLike = self._compute_log_like(modelCounts)

        return logLike

    def get_nuisance_parameters(self):
        '''
        Return a list of nuisance parameter names. Return an empty list if there
        are no nuisance parameters
        '''
        return self.nuisanceParameters.keys()

    def view_count_spectrum(self):
        '''
        View the count and background spectrum. Useful to check energy selections.

        '''
        # First plot the counts
        _ = gbm_channel_plot(self.response.ebounds[:, 0], self.response.ebounds[:, 1], self.counts_backup,
                             color='#377eb8', lw=2, alpha=1)
        ax = gbm_channel_plot(self.response.ebounds[:, 0], self.response.ebounds[:, 1], self.bkgCounts_backup,
                              color='#e41a1c', alpha=.8)
        # Now fade the non-used channels
        excluded_channel_plot(self.response.ebounds[:, 0], self.response.ebounds[:, 1], self.mask, self.counts_backup,
                              self.bkgCounts_backup, ax)

        ax.set_xlabel("Energy (keV)")
        ax.set_ylabel("Counts/keV")
        ax.set_xlim(left=self.response.ebounds[0, 0], right=self.response.ebounds[-1, 1])

    def use_intercalibration_constant(self, factorLowBound=0.9, factorHiBound=1.1):

        self.nuisanceParameters['InterCalib'].free()
        self.nuisanceParameters['InterCalib'].set_bounds(factorLowBound, factorHiBound)

        # Check that the parameter is within the provided bounds
        value = self.nuisanceParameters['InterCalib'].value

        if (value < factorLowBound):
            warnings.warn(
                "The intercalibration constant was %s, lower than the provided lower bound %s." % (
                    value, factorLowBound) +
                " Setting it equal to the lower bound")

            self.nuisanceParameters['InterCalib'].setValue(float(factorLowBound))

        if (value > factorHiBound):
            warnings.warn(
                "The intercalibration constant was %s, larger than the provided hi bound %s." % (value, factorHiBound) +
                " Setting it equal to the hi bound")

            self.nuisanceParameters['InterCalib'].setValue(float(factorHiBound))

    def fix_intercalibration_constant(self, value=None):

        if (value is not None):
            # Fixing the constant to the provided value
            self.nuisanceParameters['InterCalib'].value = float(value)

        else:

            # Do nothing, i.e., leave the constant to the value
            # it currently has
            pass

        self.nuisanceParameters['InterCalib'].fix()

    def _chisq(self, modelCounts):

        # for chisq we need sigma^2, but sigma = sqrt(counts), -> sigma^2 = counts
        # If there are zero counts then the denominator is 0, so use the model
        # variance instead

        sanitized_variance = np.where(self.counts > 0, self.counts, modelCounts)
        sanitized_variance = np.maximum(sanitized_variance, 1e-12)

        # This not really chisq, but it's a Gaussian likelihood instead. That's why there is a -1/2
        # in front of it. However, this like = -0.5 * chisq + const (we neglect the const)

        log_likes = - 0.5 * (modelCounts - self.counts) ** 2 / sanitized_variance

        return np.sum(log_likes)

    def use_chisq(self):

        self._backup_log_like = self._compute_log_like
        self._compute_log_like = self._chisq

    def restore_normal_statistic(self):

        self._compute_log_like = self._backup_log_like


class OGIPPluginPGstat(OGIPPluginBase):

    # Override the _initialSetup method just to force the existance of bkgErr
    def _initialSetup(self, mask, counts, bkgCounts, exposure, bkgErr):

        assert bkgErr is not None, "You have to provide errors on the background to use PGstat"

        super(OGIPPluginPGstat, self)._initialSetup(mask, counts, bkgCounts, exposure, bkgErr)

    def _compute_log_like(self, modelCounts):
        # This loglike assume Gaussian errors on the background and Poisson uncertainties on the
        # observed counts. It is a profile likelihood.

        MB = self.bkgCounts + modelCounts
        s2 = self.bkgErr ** 2

        b = 0.5 * (np.sqrt(MB ** 2 - 2 * s2 * (MB - 2 * self.counts) + self.bkgErr ** 4)
                   + self.bkgCounts - modelCounts - s2)

        # Now there are two branches: when the background is 0 we are in the normal situation of a pure
        # Poisson likelihood, while when the background is not zero we use the profile likelihood

        # NOTE: In the constructor we enforced that bkgErr can be 0 only when also bkgCounts = 0
        # Also it is evident from the expression above that when bkgCounts = 0 and bkgErr=0 also b=0

        # Let's do the branch with background > 0 first

        idx = self.bkgCounts > 0

        log_likes = np.empty_like(modelCounts)

        log_likes[idx] = (-(b[idx] - self.bkgCounts[idx]) ** 2 / (2 * s2[idx])
                          + self.counts[idx] * np.log(b[idx] + modelCounts[idx])
                          - b[idx] - modelCounts[idx] - logfactorial(self.counts[idx])
                          - 0.5 * log(2 * np.pi) - np.log(self.bkgErr[idx]))

        # Let's do the other branch

        nidx = ~idx

        # the 1e-100 in the log is to avoid zero divisions
        # This is the Poisson likelihood with no background
        log_likes[nidx] = self.counts[nidx] * np.log(modelCounts[nidx] + 1e-100) - modelCounts[nidx] \
                          - logfactorial(self.counts[nidx])

        return np.sum(log_likes)


class OGIPPluginCash(OGIPPluginBase):

    # Override the _initialSetup method just to force the non-existance of bkgErr
    def _initialSetup(self, mask, counts, bkgCounts, exposure, bkgErr=None):

        assert bkgErr is None, "You cannot use bkg. errors when using Cash statistic"

        super(OGIPPluginCash, self)._initialSetup(mask, counts, bkgCounts, exposure, None)

    def _compute_log_like(self, modelCounts):

        log_likes = self.counts * np.log(modelCounts + 1e-100) - modelCounts - logfactorial(self.counts)

        return np.sum(log_likes)


class OGIPPHA(object):

    def __init__(self, phafile, spectrumNumber=None, **kwargs):

        if '.root' not in phafile:

            self._init_from_FITS(phafile, spectrumNumber, **kwargs)

        else:

            self._init_from_ROOT()

    def _init_from_ROOT(self):

        raise NotImplementedError("Not yet implemented")

    def _init_from_FITS(self, phafile, spectrumNumber, **kwargs):

        self.filetype = 'observed'

        for k, v in kwargs.iteritems():

            if (k.lower() == "filetype"):

                if (v.lower() == "background"):

                    self.filetype = "background"

                elif (v.lower() == "observed"):

                    self.filetype = "observed"

                else:

                    raise RuntimeError("Unrecognized filetype keyword value")

        # Allow the use of a syntax like "mySpectrum.pha{1}" to specify the spectrum
        # number in PHA II files

        ext = os.path.splitext(phafile)[-1]

        if ('{' in ext):
            spectrumNumber = int(ext.split('{')[-1].replace('}', ''))

            phafile = phafile.split('{')[0]

        with pyfits.open(phafile) as f:

            try:

                HDUidx = f.index_of("SPECTRUM")

            except:

                raise RuntimeError("The input file %s is not in PHA format" % (phafile))

            self.spectrumNumber = spectrumNumber

            spectrum = f[HDUidx]
            data = spectrum.data
            header = spectrum.header

            # Determine if this file contains COUNTS or RATES

            self.exposure = 1

            if ("COUNTS" in data.columns.names):

                self.hasRates = False
                self.dataColumnName = "COUNTS"

            elif ("RATE" in data.columns.names):

                self.hasRates = True
                self.dataColumnName = "RATE"

                #self.exposure = header.get("EXPOSURE")

            else:

                raise RuntimeError("This file does not contain a RATE nor a COUNTS column. "
                                   "This is not a valid PHA file")

            # Determine if this is a PHA I or PHA II
            if (len(data.field(self.dataColumnName).shape) == 2):

                self.typeII = True

                if (self.spectrumNumber == None):
                    raise RuntimeError("This is a PHA Type II file. You have to provide a spectrum number")

            else:

                self.typeII = False

            # Collect informations from mandatory keywords

            keys = requiredKeywords[self.filetype]

            for k in keys:

                internalName, keyname = k.split(":")

                if (keyname in header):

                    self.__setattr__(internalName, header[keyname])

                else:

                    if (keyname in mightBeColumns[self.filetype] and self.typeII):

                        # Check if there is a column with this name

                        if (keyname in data.columns.names):

                            # This will set the exposure, among other things

                            self.__setattr__(internalName, data[keyname][self.spectrumNumber - 1])

                        else:

                            raise RuntimeError("Keyword %s is not in the header nor in the data extension. "
                                               "This file is not a proper PHA file" % keyname)

                    else:

                        # The keyword POISSERR is a special case, because even if it is missing,
                        # it is assumed to be False if there is a STAT_ERR column in the file

                        if (keyname == "POISSERR" and "STAT_ERR" in data.columns.names):

                            self.poisserr = False

                        else:

                            raise RuntimeError("Keyword %s not found. File %s is not a proper PHA "
                                               "file" % (keyname, phafile))

            # Now get the data (counts or rates) and their errors. If counts, transform them in rates
            if (self.typeII):

                # PHA II file
                if (self.hasRates):

                    self.rates = data.field(self.dataColumnName)[self.spectrumNumber - 1, :]

                    if (not self.poisserr):
                        self.ratesErrors = data.field("STAT_ERR")[self.spectrumNumber - 1, :]

                else:

                    self.rates = data.field(self.dataColumnName)[self.spectrumNumber - 1, :] / self.exposure

                    if (not self.poisserr):

                        self.ratesErrors = data.field("STAT_ERR")[self.spectrumNumber - 1, :] / self.exposure


                if ("SYS_ERR" in data.columns.names):

                    self.sysErrors = data.field("SYS_ERR")[self.spectrumNumber - 1, :]
                else:

                    self.sysErrors = np.zeros(self.rates.shape)


            elif (self.typeII == False):

                # PHA 1 file
                if (self.hasRates):

                    self.rates = data.field(self.dataColumnName)

                    if (not self.poisserr):
                        self.ratesErrors = data.field("STAT_ERR")
                else:

                    self.rates = data.field(self.dataColumnName) / self.exposure

                    if (not self.poisserr):
                        self.ratesErrors = data.field("STAT_ERR") / self.exposure

                if ("SYS_ERR" in data.columns.names):

                    self.sysErrors = data.field("SYS_ERR")

                else:

                    self.sysErrors = np.zeros(self.rates.shape)

    def getRates(self):
        return self.rates

    def getRatesErrors(self):

        assert self.poisserr == False, "Cannot request errors on rates for a Poisson spectrum"

        return self.ratesErrors

    def getSysErrors(self):
        return self.sysErrors

    def getExposure(self):
        return self.exposure


class Response(object):
    def __init__(self, rspfile, arffile=None):

        rspNumber = 1

        # Check if we are dealing with a .rsp2 file (containing more than
        # one response). This is checked by looking for the syntax
        # [responseFile]{[responseNumber]}

        if ('{' in rspfile):
            tokens = rspfile.split("{")
            rspfile = tokens[0]
            rspNumber = int(tokens[-1].split('}')[0].replace(" ", ""))

        # Read the response
        with pyfits.open(rspfile) as f:

            try:

                # GBM typical response
                data = f['MATRIX', rspNumber].data

            except:
                # Other detectors might use the SPECRESP MATRIX name instead
                # Note that here we are not catching any exception, because
                # we have to fail if we cannot read the matrix

                data = f['SPECRESP MATRIX', rspNumber].data

            # Sometimes .rsp files contains a weird format featuring variable-length
            # arrays. Historically those confuse pyfits quite a lot, so we ensure
            # to transform them into standard numpy matrices to avoid issues

            self.matrix = variableToMatrix(data.field('MATRIX'))

            self.ebounds = np.vstack([f['EBOUNDS'].data.field("E_MIN"),
                                      f['EBOUNDS'].data.field("E_MAX")]).T

            self.mc_channels = np.vstack([data.field("ENERG_LO"),
                                          data.field("ENERG_HI")]).T

            # Now let's see if we have a ARF, if yes, read it

            if arffile is not None:

                with pyfits.open(arffile) as f:

                    data = f['SPECRESP'].data

                arf = data.field('SPECRESP')

                # Check that arf and rmf have same dimensions

                if arf.shape[0] != self.matrix.shape[1]:
                    raise IOError("The ARF and the RMF file does not have the same number of channels")

                # Check that the ENERG_LO and ENERG_HI for the RMF and the ARF
                # are the same

                arf_mc_channels = np.vstack([data.field("ENERG_LO"),
                                             data.field("ENERG_HI")]).T

                # Declare the mc channels different if they differ by more than
                # 1%

                idx = (self.mc_channels > 0)

                diff = (self.mc_channels[idx] - arf_mc_channels[idx]) / self.mc_channels[idx]

                if diff.max() > 0.01:
                    raise IOError("The ARF and the RMF have one or more MC channels which differ by more than 1%")

                # Multiply ARF and RMF

                self.matrix = self.matrix * arf

    def set_function(self, differentialFunction, integralFunction=None):
        '''
        Set the function to be used for the convolution
        '''

        self.differentialFunction = differentialFunction

        self.integralFunction = integralFunction

    def convolve(self):

        trueFluxes = self.integralFunction(self.mc_channels[:, 0],
                                           self.mc_channels[:, 1])

        # Sometimes some channels have 0 lenths, or maybe they start at 0, where
        # many functions (like a power law) are not defined. In the response these
        # channels have usually a 0, but unfortunately for a computer
        # inf * zero != zero. Thus, let's force this. We avoid checking this situation
        # in details because this would have a HUGE hit on performances

        idx = np.isfinite(trueFluxes)
        trueFluxes[~idx] = 0

        foldedCounts = np.dot(trueFluxes, self.matrix.T)

        return foldedCounts

    def getCountsVector(self, e1, e2):

        trueFluxes = self.integralFunction(self.mc_channels[:, 0], self.mc_channels[:, 1])

        return trueFluxes

    def energy_to_channel(self, energy):

        '''Finds the channel containing the provided energy.
        NOTE: returns the channel index (starting at zero),
        not the channel number (likely starting from 1)'''

        # Get the index of the first ebounds upper bound larger than energy

        try:

            idx = next(idx for idx,
                               value in enumerate(self.ebounds[:, 1])
                       if value >= energy)

        except StopIteration:

            # No values above the given energy, return the last channel
            return self.ebounds[:, 1].shape[0]

        return idx


def variableToMatrix(variableLengthMatrix):
    '''This take a variable length array and return it in a
    properly formed constant length array, to avoid some pyfits obscure bugs'''

    nrows = len(variableLengthMatrix)
    ncolumns = max([len(elem) for elem in variableLengthMatrix])
    matrix = np.zeros([ncolumns, nrows])

    for i in range(nrows):
        for j in range(ncolumns):

            try:

                matrix[j, i] = variableLengthMatrix[i][j]

            except:

                # This happens when the i,j element does not exist, which is not an error
                # We will just leave it to zero
                pass

    return matrix


def gbm_channel_plot(chan_min, chan_max, counts, **keywords):
    chans = np.array(zip(chan_min, chan_max))
    width = chan_max - chan_min
    fig = plt.figure(666)
    ax = fig.add_subplot(111)
    step_plot(chans, counts / width, ax, **keywords)
    ax.set_xscale('log')
    ax.set_yscale('log')

    return ax


def excluded_channel_plot(chan_min, chan_max, mask, counts, bkg, ax):
    # Figure out the best limit
    chans = np.array(zip(chan_min, chan_max))
    width = chan_max - chan_min

    top = max([max(bkg / width), max(counts / width)])
    top = top + top * .5
    bottom = min([min(bkg / width), min(counts / width)])
    bottom = bottom - bottom * .2

    # Find the contiguous regions
    slices = slice_disjoint((~mask).nonzero()[0])

    for region in slices:
        ax.fill_between([chan_min[region[0]], chan_max[region[1]]],
                        bottom,
                        top,
                        color='k',
                        alpha=.5)

    ax.set_ylim(bottom, top)


def slice_disjoint(arr):
    slices = []
    startSlice = 0
    counter = 0
    for i in range(len(arr) - 1):
        if arr[i + 1] > arr[i] + 1:
            endSlice = arr[i]
            slices.append([startSlice, endSlice])
            startSlice = arr[i + 1]
            counter += 1
    if counter == 0:
        return [[arr[0], arr[-1]]]
    if endSlice != arr[-1]:
        slices.append([startSlice, arr[-1]])
    return slices


def gbm_light_curve_plot(time_bins, cnts, bkg, width, selection):
    fig = plt.figure(777)
    ax = fig.add_subplot(111)

    maxCnts = max(cnts / width)
    top = maxCnts + maxCnts * .2
    minCnts = min(cnts[cnts > 0] / width)
    bottom = minCnts - minCnts * .2
    mean_time = map(np.mean, time_bins)

    step_plot(time_bins, cnts / width, ax, color='#8da0cb')

    ax.plot(mean_time, bkg, '#66c2a5', lw=2.)

    ax.fill_between(selection, bottom, top, color="#fc8d62", alpha=.4)

    ax.set_xlabel("Time (s)")
    ax.set_ylabel("Rate (cnts/s)")
    ax.set_ylim(bottom, top)


import collections
import math

import numpy as np
import pandas as pd
import numdifftools as nd
from iminuit import Minuit

from astromodels import SettingOutOfBounds

from threeML.io.progress_bar import ProgressBar
import scipy.optimize

from threeML.exceptions.custom_exceptions import custom_warnings
from threeML.utils.differentiation import get_hessian, ParameterOnBoundary

# Set the warnings to be issued always for this module

custom_warnings.simplefilter("always", RuntimeWarning)


try:

    import ROOT

except ImportError:

    has_ROOT = False

else:

    has_ROOT = True

try:

    import pyOpt

except ImportError:

    has_pyOpt = False

else:

    has_pyOpt = True

# Special constants
FIT_FAILED = 1e12


# Define a bunch of custom exceptions relevant for what is being accomplished here

class CannotComputeCovariance(RuntimeWarning):
    pass


class CannotComputeErrors(RuntimeWarning):
    pass


class MINOSFailed(Exception):
    pass


class ParameterIsNotFree(Exception):
    pass


class MinimizerNotAvailable(Exception):
    pass


class BetterMinimumDuringProfiling(RuntimeWarning):
    pass

# This will contain the available minimizers

_minimizers = {}


def get_minimizer(minimizer_type):
    """
    Return the requested minimizer *class* (not instance)

    :param minimizer_type: MINUIT, ROOT, or PYOPT
    :param minimizer_algorithm: algorithm (optional, use only for PYOPT)
    :return: the class (i.e., the type) for the requested minimizer
    """

    try:

        return _minimizers[minimizer_type]

    except KeyError:

        raise MinimizerNotAvailable("Minimizer %s is not available on your system" % minimizer_type)


class FunctionWrapper(object):

    def __init__(self, function, all_parameters, fixed_parameters):
        """

        :param function:
        :param all_parameters:
        :param fixed_parameters: list of fixed parameters
        """
        self._function = function

        self._all_parameters = all_parameters

        self._fixed_parameters_values = np.zeros(len(fixed_parameters))
        self._fixed_parameters_names = fixed_parameters

        self._indexes_of_fixed_par = np.zeros(len(self._all_parameters), bool)

        for i, parameter_name in enumerate(self._fixed_parameters_names):

            this_index = self._all_parameters.keys().index(parameter_name)

            self._indexes_of_fixed_par[this_index] = True

        self._all_values = np.zeros(len(self._all_parameters))

    def set_fixed_values(self, new_fixed_values):

        # A use [:] so there is an implicit check on the right size of new_fixed_values

        self._fixed_parameters_values[:] = new_fixed_values

    def __call__(self, *trial_values):

        self._all_values[self._indexes_of_fixed_par] = self._fixed_parameters_values
        self._all_values[~self._indexes_of_fixed_par] = trial_values

        return self._function(*self._all_values)


class ProfileLikelihood(object):

    def __init__(self, minimizer_instance, fixed_parameters):

        self._original_minimizer = minimizer_instance

        self._fixed_parameters = fixed_parameters

        assert len(self._fixed_parameters) <= 2, "Can handle only one or two fixed parameters"

        # Get some info from the original minimizer

        self._function = self._original_minimizer.function

        self._all_parameters = self._original_minimizer.parameters

        ftol = self._original_minimizer.ftol

        # Create a copy of the dictionary of parameters

        free_parameters = collections.OrderedDict(self._all_parameters)

        # Remove the fixed ones

        for parameter_name in fixed_parameters:

            free_parameters.pop(parameter_name)

        # Now compute how many free parameters we have

        self._n_free_parameters = len(free_parameters)

        if self._n_free_parameters > 0:

            self._wrapper = FunctionWrapper(self._function,
                                            self._all_parameters,
                                            self._fixed_parameters)

            # Create a copy of the optimizer with the new parameters (i.e., one or two
            # parameters fixed to their current values)

            self._optimizer = type(self._original_minimizer)(self._wrapper, free_parameters, ftol, verbosity=0)

            if self._original_minimizer.algorithm_name is not None:

                self._optimizer.set_algorithm(self._original_minimizer.algorithm_name)

        else:

            # Special case when there are no free parameters after fixing the requested ones
            # There is no profiling necessary here

            self._wrapper = None
            self._optimizer = None

    def step(self, steps1, steps2=None):

        if steps2 is not None:

            assert len(self._fixed_parameters) == 2, "Cannot step in 2d if you fix only one parameter"

            # Find out if the user is giving flipped steps (i.e. param_1 is after param_2 in the
            # parameters dictionary)

            param_1_name = self._fixed_parameters[0]
            param_1_idx = self._all_parameters.keys().index(param_1_name)

            param_2_name = self._fixed_parameters[1]
            param_2_idx = self._all_parameters.keys().index(param_2_name)

            if param_1_idx > param_2_idx:

                # Switch steps

                swap = steps1
                steps1 = steps2
                steps2 = swap

                results = self._step2d(steps1, steps2).T

            else:

                results = self._step2d(steps1, steps2)

            return results

        else:

            assert len(self._fixed_parameters) == 1, "You cannot step in 1d if you fix 2 parameters"

            return self._step1d(steps1)

    def __call__(self, values):

        self._wrapper.set_fixed_values(values)

        _, this_log_like = self._optimizer.minimize(compute_covar=False)

        return this_log_like

    def _step1d(self, steps1):

        log_likes = np.zeros_like(steps1)

        p = ProgressBar(len(steps1))

        for i, step in enumerate(steps1):

            if self._n_free_parameters > 0:

                # Profile out the free parameters

                self._wrapper.set_fixed_values(step)

                _, this_log_like = self._optimizer.minimize(compute_covar=False)

            else:

                # No free parameters, just compute the likelihood

                this_log_like = self._function(step)

            log_likes[i] = this_log_like

            p.increase()

        return log_likes

    def _step2d(self, steps1, steps2):

        log_likes = np.zeros((len(steps1), len(steps2)))

        p = ProgressBar(len(steps1) * len(steps2))

        for i, step1 in enumerate(steps1):

            for j,step2 in enumerate(steps2):

                if self._n_free_parameters > 0:

                    # Profile out the free parameters

                    self._wrapper.set_fixed_values([step1, step2])

                    _, this_log_like = self._optimizer.minimize(compute_covar=False)

                else:

                    # No free parameters, just compute the likelihood

                    this_log_like = self._function(step1, step2)

                log_likes[i,j] = this_log_like

                p.increase()

        return log_likes


class Minimizer(object):

    def __init__(self, function, parameters, ftol=1e-3, verbosity=1):
        """

        :param function: function to be minimized
        :param parameters: ordered dictionary of the FREE parameters in the fit. The order must be the same as
               in the calling sequence of the function to be minimized.
        :param ftol: fractional tolerance to be used in the fit
        :param verbosity: control the verbosity of the output
        :return:
        """

        self.function = function
        self.parameters = parameters
        self.Npar = len(self.parameters.keys())
        self.ftol = ftol
        self.verbosity = verbosity

        self._setup()

        self._fit_results = None
        self._covariance_matrix = None
        self._correlation_matrix = None

        self._algorithm_name = None
        self._m_log_like_minimum = None

    def _setup(self):

        raise NotImplementedError("You have to implement this.")

    @property
    def algorithm_name(self):

        return self._algorithm_name

    def minimize(self):
        raise NotImplemented("This is the method of the base class. Must be implemented by the actual minimizer")

    def set_algorithm(self, algorithm):

        raise NotImplementedError("Must be implemented by the actual minimizer if it provides more than one algorithm")

    def _store_fit_results(self, best_fit_values, m_log_like_minimum, covariance_matrix=None):

        self._m_log_like_minimum = m_log_like_minimum

        # Create a pandas DataFrame with the fit results

        values = collections.OrderedDict()
        errors = collections.OrderedDict()

        for i in range(self.Npar):

            name = self.parameters.keys()[i]

            value = best_fit_values[i]

            if covariance_matrix is not None:

                element = covariance_matrix[i,i]

                if element > 0:

                    error = math.sqrt(covariance_matrix[i, i])

                else:

                    custom_warnings.warn("Negative element on diagonal of covariance matrix", CannotComputeErrors)

                    error = np.nan

            else:

                error = np.nan

            values[name] = value
            errors[name] = error

        data = collections.OrderedDict()
        data['value'] = pd.Series(values)
        data['error'] = pd.Series(errors)

        self._fit_results = pd.DataFrame(data)
        self._covariance_matrix = covariance_matrix

        # Compute correlation matrix

        self._correlation_matrix = np.zeros_like(self._covariance_matrix)

        if covariance_matrix is not None:

            for i in range(self.Npar):

                variance_i = self._covariance_matrix[i,i]

                for j in range(self.Npar):

                    variance_j = self._covariance_matrix[j,j]

                    if variance_i * variance_j > 0:

                        self._correlation_matrix[i,j] = self._covariance_matrix[i,j] / (math.sqrt(variance_i * variance_j))

                    else:

                        # We already issued a warning about this, so let's quietly fail

                        self._correlation_matrix[i, j] = np.nan

    @property
    def fit_results(self):

        return self._fit_results

    @property
    def covariance_matrix(self):

        return self._covariance_matrix

    @property
    def correlation_matrix(self):

        return self._correlation_matrix

    def restore_best_fit(self):
        """
        Reset all the parameters to their best fit value (from the last run fit)

        :return: none
        """

        best_fit_values = self._fit_results['value'].values

        for parameter_name, best_fit_value in zip(self.parameters.keys(), best_fit_values):

            self.parameters[parameter_name].value = best_fit_value

    def _compute_covariance_matrix(self, best_fit_values):
        """
        This function compute the approximate covariance matrix as the inverse of the Hessian matrix,
        which is the matrix of second derivatives of the likelihood function with respect to
        the parameters.

        The sqrt of the diagonal of the result is an accurate estimate of the errors only if the
        log.likelihood is parabolic in the neighborhood of the minimum.

        Derivatives are computed numerically.

        :return: the covariance matrix
        """

        minima = np.array(map(lambda parameter:parameter.min_value, self.parameters.values()))
        maxima = np.array(map(lambda parameter: parameter.max_value, self.parameters.values()))

        try:

            hessian_matrix = get_hessian(self.function, best_fit_values, minima, maxima)

        except ParameterOnBoundary:

            custom_warnings.warn("One or more of the parameters are at their boundaries. Cannot compute covariance and"
                                 " errors", CannotComputeCovariance)

            n_dim = len(best_fit_values)

            return np.zeros((n_dim,n_dim)) * np.nan

        # Invert it to get the covariance matrix

        covariance_matrix = np.linalg.inv(hessian_matrix)

        # Now check that the covariance matrix is semi-positive definite (it must be unless
        # there have been numerical problems, which can happen when some parameter is unconstrained)

        # The fastest way is to try and compute the Cholesky decomposition, which
        # works only if the matrix is positive definite

        try:

            _ = np.linalg.cholesky(covariance_matrix)

        except:

            custom_warnings.warn("Covariance matrix is NOT semi-positive definite. Cannot estimate errors. This can "
                                 "happen for many reasons, the most common being one or more unconstrained parameters",
                                 CannotComputeCovariance)

        return covariance_matrix

    def _get_error(self, parameter_name, target_delta_log_like, sign=-1):

        # Since the procedure might find a better minimum, we can repeat it
        # up to a maximum of 10 times

        repeats = 0

        while repeats < 10:

            # Let's start optimistic...

            repeat = False

            repeats += 1

            self.restore_best_fit()

            best_fit_value = self.parameters[parameter_name].value

            if sign == -1:

                extreme_allowed = self.parameters[parameter_name].min_value

            else:

                extreme_allowed = self.parameters[parameter_name].max_value

            # If the parameter has no boundary in the direction we are sampling, put a hard limit on
            # 10 times the current value (to avoid looping forever)

            if extreme_allowed is None:

                extreme_allowed = best_fit_value + sign * 10 * abs(best_fit_value)

            # We need to look for a value for the parameter where the difference between the minimum of the
            # log-likelihood and the likelihood for that value differs by more than target_delta_log_likelihood.
            # This is needed by the root-finding procedure, which needs to know an interval where the biased likelihood
            # function (see below) changes sign

            trials = best_fit_value + sign * np.linspace(0.1, 0.9, 9) * abs(best_fit_value)

            trials = np.append(trials, extreme_allowed)

            # Make sure we don't go below the allowed minimum or above the allowed maximum

            if sign == -1:

                np.clip(trials, extreme_allowed, np.inf, trials)

            else:

                np.clip(trials, -np.inf, extreme_allowed, trials)

            # There might be more than one value which was below the minimum (or above the maximum), so let's
            # take only unique elements

            trials = np.unique(trials)

            trials.sort()

            if sign == -1:

                trials = trials[::-1]

            # At this point we have a certain number of unique trials which always
            # contain the allowed minimum (or maximum)

            minimum_bound = None
            maximum_bound = None

            # Instance the profile likelihood function
            pl = ProfileLikelihood(self, [parameter_name])

            for i, trial in enumerate(trials):

                this_log_like = pl([trial])

                delta = this_log_like - self._m_log_like_minimum

                if delta < -0.1:

                    custom_warnings.warn("Found a better minimum (%.2f) for %s = %s during error "
                                         "computation." % (this_log_like, parameter_name, trial),
                                         BetterMinimumDuringProfiling)

                    xs = map(lambda x:x.value, self.parameters.values())

                    self._store_fit_results(xs, this_log_like, None)

                    repeat = True

                    break

                if delta > target_delta_log_like:

                    bound1 = trial

                    if i > 0:

                        bound2 = trials[i-1]

                    else:

                        bound2 = best_fit_value

                    minimum_bound = min(bound1, bound2)
                    maximum_bound = max(bound1, bound2)

                    repeat = False

                    break

            if repeat:

                # We found a better minimum, restart from scratch

                custom_warnings.warn("Restarting search...", RuntimeWarning)

                continue

            if minimum_bound is None:

                # Cannot find error in this direction (it's probably outside the allowed boundaries)
                custom_warnings.warn("Cannot find lower boundary for parameter %s" % parameter_name, CannotComputeErrors)

                error = np.nan
                break

            else:

                # Define the "biased likelihood", since brenq only finds zeros of function

                biased_likelihood = lambda x: pl(x) - self._m_log_like_minimum - target_delta_log_like

                if sign == -1:

                    precise_bound = scipy.optimize.brentq(biased_likelihood, minimum_bound,
                                                          maximum_bound, xtol=1e-5, maxiter=1000)

                else:

                    precise_bound = scipy.optimize.brentq(biased_likelihood, minimum_bound,
                                                          maximum_bound, xtol=1e-5, maxiter=1000)

                error = precise_bound - best_fit_value

                break

        return error

    def get_errors(self):
        """
        Compute asymmetric errors using the profile likelihood method (slow, but accurate).

        :return: a list with asymmetric errors for each parameter
        """

        # TODO: options for other significance levels

        target_delta_log_like = 0.5

        self.restore_best_fit()

        p = ProgressBar(2 * len(self.parameters))

        errors = collections.OrderedDict()

        for parameter_name in self.parameters:

            negative_error = self._get_error(parameter_name, target_delta_log_like, -1)

            p.increase()

            positive_error = self._get_error(parameter_name, target_delta_log_like, +1)

            p.increase()

            errors[parameter_name] = (negative_error, positive_error)

        return errors

    def contours(self, param_1, param_1_minimum, param_1_maximum, param_1_n_steps,
                         param_2=None, param_2_minimum=None, param_2_maximum=None, param_2_n_steps=None,
                         progress=True, **options):

            """
            Generate confidence contours for the given parameters by stepping for the given number of steps between
            the given boundaries. Call it specifying only source_1, param_1, param_1_minimum and param_1_maximum to
            generate the profile of the likelihood for parameter 1. Specify all parameters to obtain instead a 2d
            contour of param_1 vs param_2

            :param param_1: name of the first parameter
            :param param_1_minimum: lower bound for the range for the first parameter
            :param param_1_maximum: upper bound for the range for the first parameter
            :param param_1_n_steps: number of steps for the first parameter
            :param param_2: name of the second parameter
            :param param_2_minimum: lower bound for the range for the second parameter
            :param param_2_maximum: upper bound for the range for the second parameter
            :param param_2_n_steps: number of steps for the second parameter
            :param progress: (True or False) whether to display progress or not
            :param log: by default the steps are taken linearly. With this optional parameter you can provide a tuple of
            booleans which specify whether the steps are to be taken logarithmically. For example,
            'log=(True,False)' specify that the steps for the first parameter are to be taken logarithmically, while they
            are linear for the second parameter. If you are generating the profile for only one parameter, you can specify
             'log=(True,)' or 'log=(False,)' (optional)
            :param: parallel: whether to use or not parallel computation (default:False)
            :return: a : an array corresponding to the steps for the first parameter
                     b : an array corresponding to the steps for the second parameter (or None if stepping only in one
                     direction)
                     contour : a matrix of size param_1_steps x param_2_steps containing the value of the function at the
                     corresponding points in the grid. If param_2_steps is None (only one parameter), then this reduces to
                     an array of size param_1_steps.
            """

            # Figure out if we are making a 1d or a 2d contour

            if param_2 is None:

                n_dimensions = 1
                fixed_parameters = [param_1]

            else:

                n_dimensions = 2
                fixed_parameters = [param_1, param_2]

            # Check the options

            p1log = False
            p2log = False

            if 'log' in options.keys():

                assert len(options['log']) == n_dimensions, ("When specifying the 'log' option you have to provide a " +
                                                             "boolean for each dimension you are stepping on.")

                p1log = bool(options['log'][0])

                if param_2 is not None:

                    p2log = bool(options['log'][1])

            # Generate the steps

            if p1log:

                param_1_steps = np.logspace(math.log10(param_1_minimum), math.log10(param_1_maximum),
                                            param_1_n_steps)

            else:

                param_1_steps = np.linspace(param_1_minimum, param_1_maximum,
                                            param_1_n_steps)

            if n_dimensions == 2:

                if p2log:

                    param_2_steps = np.logspace(math.log10(param_2_minimum), math.log10(param_2_maximum),
                                                param_2_n_steps)

                else:

                    param_2_steps = np.linspace(param_2_minimum, param_2_maximum,
                                                param_2_n_steps)

            else:

                # Only one parameter to step through
                # Put param_2_steps as nan so that the worker can realize that it does not have
                # to step through it

                param_2_steps = np.array([np.nan])

            # Define the worker which will compute the value of the function at a given point in the grid

            # Restore best fit

            if self.fit_results is not None:

                self.restore_best_fit()

            else:

                custom_warnings.warn("No best fit to restore before contours computation. "
                                     "Perform the fit before running contours to remove this warnings.")

            pr = ProfileLikelihood(self, fixed_parameters)

            if n_dimensions == 1:

                results = pr.step(param_1_steps)

            else:

                results = pr.step(param_1_steps, param_2_steps)

            # Return results

            return param_1_steps, param_2_steps, np.array(results).reshape((param_1_steps.shape[0],
                                                                            param_2_steps.shape[0]))

    # def print_fit_results(self):
    #     """
    #     Display the results of the last minimization.
    #
    #     :return: (none)
    #     """
    #
    #     data = []
    #
    #     # Also store the maximum length to decide the length for the line
    #
    #     name_length = 0
    #
    #     for parameter_name in self._fit_results.index.values:
    #
    #         value = self._fit_results.at[parameter_name, 'value']
    #
    #         error = self._fit_results.at[parameter_name, 'error']
    #
    #         # Format the value and the error with sensible significant
    #         # numbers
    #         x = uncertainties.ufloat(value, error)
    #
    #         # Add some space around the +/- sign
    #
    #         rep = x.__str__().replace("+/-", " +/- ")
    #
    #         data.append([parameter_name, rep, self.parameters[parameter_name].unit])
    #
    #         if len(parameter_name) > name_length:
    #
    #             name_length = len(parameter_name)
    #
    #     table = Table(rows=data,
    #                   names=["Name", "Best fit value", "Unit"],
    #                   dtype=('S%i' % name_length, str, str))
    #
    #     display(table)
    #
    #     print("\nNOTE: errors on parameters are approximate. Use get_errors().\n")


# This is a function to add a method to a class
# We will need it in the MinuitMinimizer

def add_method(self, method, name=None):
    if name is None:
        name = method.func_name

    setattr(self.__class__, name, method)


class MinuitMinimizer(Minimizer):

    # NOTE: this class is built to be able to work both with iMinuit and with a boost interface to SEAL
    # minuit, i.e., it does not rely on functionality that iMinuit provides which is not of the original
    # minuit. This makes the implementation a little bit more cumbersome, but more adaptable if we want
    # to switch back to the bare bone SEAL minuit

    def __init__(self, function, parameters, ftol=1, verbosity=0):

        super(MinuitMinimizer, self).__init__(function, parameters, ftol, verbosity)

    def _setup(self):

        # Prepare the dictionary for the parameters which will be used by iminuit

        iminuit_init_parameters = {}

        # List of variable names that will be used for iminuit.

        variable_names_for_iminuit = []

        # NOTE: we use the scaled_ versions of value, min_value and max_value because they don't have
        # units, and hence they are much faster to set and retrieve. These are indeed introduced by
        # astromodels to be used for computing-intensive situations like fitting

        for k, par in self.parameters.iteritems():

            current_name = self._parameter_name_to_minuit_name(k)

            variable_names_for_iminuit.append(current_name)

            # Initial value
            iminuit_init_parameters['%s' % current_name] = par.value

            # Initial delta
            iminuit_init_parameters['error_%s' % current_name] = par.delta

            # Limits
            iminuit_init_parameters['limit_%s' % current_name] = (par.min_value, par.max_value)

            # This is useless, since all parameters here are free,
            # but do it anyway for clarity
            iminuit_init_parameters['fix_%s' % current_name] = False

        # This is to tell Minuit that we are dealing with likelihoods,
        # not chi square
        iminuit_init_parameters['errordef'] = 0.5

        iminuit_init_parameters['print_level'] = self.verbosity

        # We need to make a function with the parameters as explicit
        # variables in the calling sequence, so that Minuit will be able
        # to probe the parameter's names
        var_spelled_out = ",".join(variable_names_for_iminuit)

        # A dictionary to keep a way to convert from var. name to
        # variable position in the function calling sequence
        # (will use this in contours)

        self.name_to_position = {k: i for i, k in enumerate(variable_names_for_iminuit)}

        # Write and compile the code for such function

        code = 'def _f(self, %s):\n  return self.function(%s)' % (var_spelled_out, var_spelled_out)
        exec code

        # Add the function just created as a method of the class
        # so it will be able to use the 'self' pointer
        add_method(self, _f, "_f")

        # Finally we can instance the Minuit class
        self.minuit = Minuit(self._f, **iminuit_init_parameters)

        self.minuit.tol = self.ftol  # ftol

        try:

            self.minuit.up = 0.5  # This is a likelihood

        except AttributeError:

            # iMinuit uses errodef, not up

            self.minuit.errordef = 0.5

        self.minuit.strategy = 0  # More accurate

        self._best_fit_parameters = None
        self._function_minimum_value = None

    @staticmethod
    def _parameter_name_to_minuit_name(parameter):
        """
        Translate the name of the parameter to the format accepted by Minuit

        :param parameter: the parameter name, of the form source.component.shape.parname
        :return: a minuit-friendly name for the parameter, such as source_component_shape_parname
        """

        return parameter.replace(".", "_")

    def _migrad_has_converged(self):

        # In the MINUIT manual this is the condition for MIGRAD to have converged
        # 0.002 * tolerance * UPERROR (which is 0.5 for likelihood)

        return self.minuit.edm <= 0.002 * self.minuit.tol * 0.5

    def _run_migrad(self, trials=10):

        # Repeat Migrad up to trials times, until it converges

        minimum = None

        for i in range(trials):

            self.minuit.migrad()

            minimum = self.minuit.fval

            if self._migrad_has_converged():

                # Converged
                break

            else:

                # Try again
                continue

        return minimum

    # Override this because minuit uses different names
    def restore_best_fit(self):
        """
        Set the parameters back to their best fit value

        :return: none
        """

        super(MinuitMinimizer, self).restore_best_fit()

        for k, par in self.parameters.iteritems():

            minuit_name = self._parameter_name_to_minuit_name(k)

            self.minuit.values[minuit_name] = par.value

    def minimize(self, compute_covar=True):
        """
        Minimize the function using MIGRAD

        :param compute_covar: whether to compute the covariance (and error estimates) or not
        :return: best_fit: a dictionary containing the parameters at their best fit values
                 function_minimum : the value for the function at the minimum

                 NOTE: if the minimization fails, the dictionary will be empty and the function_minimum will be set
                 to minimization.FIT_FAILED
        """

        minimum = self._run_migrad(10)

        if not self._migrad_has_converged():

            print("\nMIGRAD did not converge in 10 trials.")

            return collections.OrderedDict(), FIT_FAILED

        else:

            # Get the best fit values

            best_fit_values = []

            for k, par in self.parameters.iteritems():

                minuit_name = self._parameter_name_to_minuit_name(k)

                best_fit_values.append(self.minuit.values[minuit_name])

            # Get covariance matrix

            if compute_covar:

                covariance = self._compute_covariance_matrix(best_fit_values)

            else:

                covariance = None

            # Now store the results

            self._store_fit_results(best_fit_values, minimum, covariance)

            return best_fit_values, minimum

    # Override the default _compute_covariance_matrix
    def _compute_covariance_matrix(self, best_fit_values):

        self.minuit.hesse()

        covariance = np.array(self.minuit.matrix(correlation=False))

        return covariance

    # def print_fit_results(self):
    #     """
    #     Display the results of the last minimization.
    #
    #     :return: (none)
    #     """
    #
    #     # Restore the best fit values, in case something has changed
    #     self._restore_best_fit()
    #
    #     # I do not use the print_param facility in iminuit because
    #     # it does not work well with console output, since it fails
    #     # to autoprobe that it is actually run in a console and uses
    #     # the HTML backend instead
    #
    #     # Create a list of strings to print
    #
    #     data = []
    #
    #     # Also store the maximum length to decide the length for the line
    #
    #     name_length = 0
    #
    #     for k, v in self.parameters.iteritems():
    #
    #         minuit_name = self._parameter_name_to_minuit_name(k)
    #
    #         # Format the value and the error with sensible significant
    #         # numbers
    #         x = uncertainties.ufloat(v.value, self.minuit.errors[minuit_name])
    #
    #         # Add some space around the +/- sign
    #
    #         rep = x.__str__().replace("+/-", " +/- ")
    #
    #         data.append([k, rep, v.unit])
    #
    #         if len(k) > name_length:
    #             name_length = len(k)
    #
    #     table = Table(rows=data,
    #                   names=["Name", "Value", "Unit"],
    #                   dtype=('S%i' % name_length, str, str))
    #
    #     display(table)
    #
    #     print("\nNOTE: errors on parameters are approximate. Use get_errors().\n")

    # Override the default _compute_covariance_matrix

    # def print_correlation_matrix(self):
    #     """
    #     Display the current correlation matrix
    #     :return: (none)
    #     """
    #
    #     # Print a custom covariance matrix because iminuit does
    #     # not guess correctly the frontend when 3ML is used
    #     # from terminal
    #
    #     cov = self.minuit.covariance
    #
    #     if cov is None:
    #         raise CannotComputeCovariance("Cannot compute covariance numerically. This usually means that there are " +
    #                                       " unconstrained parameters. Fix those or reduce their allowed range, or " +
    #                                       "use a simpler model.")
    #
    #     # Get list of parameters
    #
    #     keys = self.parameters.keys()
    #
    #     # Convert them to the format for iminuit
    #
    #     minuit_names = map(lambda k: self._parameter_name_to_minuit_name(k), keys)
    #
    #     # Accumulate rows and compute the maximum length of the names
    #
    #     data = []
    #     length_of_names = 0
    #
    #     for key1, name1 in zip(keys, minuit_names):
    #
    #         if len(name1) > length_of_names:
    #             length_of_names = len(name1)
    #
    #         this_row = []
    #
    #         for key2, name2 in zip(keys, minuit_names):
    #             # Compute correlation between parameter key1 and key2
    #
    #             corr = cov[(name1, name2)] / (math.sqrt(cov[(name1, name1)]) * math.sqrt(cov[(name2, name2)]))
    #
    #             this_row.append(corr)
    #
    #         data.append(this_row)
    #
    #     # Prepare the dtypes for the matrix
    #
    #     dtypes = map(lambda x: float, minuit_names)
    #
    #     # Column names are the parameter names
    #
    #     cols = keys
    #
    #     # Finally generate the matrix with the names
    #
    #     table = NumericMatrix(rows=data,
    #                           names=cols,
    #                           dtype=dtypes)
    #
    #     # Customize the format to avoid too many digits
    #
    #     for col in table.colnames:
    #         table[col].format = '2.2f'
    #
    #     display(table)

    def get_errors(self):
        """
        Compute asymmetric errors using MINOS (slow, but accurate) and print them.

        NOTE: this should be called immediately after the minimize() method

        :return: a dictionary containing the asymmetric errors for each parameter.
        """

        self.restore_best_fit()

        if not self._migrad_has_converged():
            raise CannotComputeErrors("MIGRAD results not valid, cannot compute errors. Did you run the fit first ?")

        try:

            self.minuit.minos()

        except:

            raise

        # except:
        #
        #     raise MINOSFailed("MINOS has failed. This usually means that the fit is very difficult, for example "
        #                       "because of high correlation between parameters. Check the correlation matrix printed"
        #                       "in the fit step, and check contour plots with getContours(). If you are using a "
        #                       "user-defined model, you can also try to "
        #                       "reformulate your model with less correlated parameters.")

        # Make a list for the results

        errors = collections.OrderedDict()

        for k, par in self.parameters.iteritems():

            minuit_name = self._parameter_name_to_minuit_name(k)

            minus_error = self.minuit.merrors[(minuit_name, -1)]
            plus_error = self.minuit.merrors[(minuit_name, 1)]

            errors[k] = ((minus_error,plus_error))

        return errors


# Add Minuit to the available minimizers
_minimizers["MINUIT"] = MinuitMinimizer


if has_pyOpt:

    class PyOptWrapper(object):

        # This is needed by pyopt

        __name__ = "Likelihood"

        def __init__(self, function, dimensions):

            self._function = function
            self._dimensions = dimensions

        def __call__(self, x):

            new_args = map(lambda i: x[i], range(self._dimensions))

            try:

                f = self._function(*new_args)

            except SettingOutOfBounds:

                f = FIT_FAILED

            if f == FIT_FAILED:

                # Likelihood gave nan or other problems, we are likely in a forbidden
                # space

                fail = 1

            else:

                # Likelihood computation successful

                fail = 0

            # The empty list is for the constraints vector. It is empty
            # because this is a unconstrained problem, where unconstrained means
            # there are no additional conditions on top of the boundaries for the
            # parameters (if any)

            return f, [], fail


    def get_pyopt_available_algorithms():
        """
        Returns a dictionary with the name of the optimizers as key and the relative class as value

        :return: dictionary
        """

        optimizers = {}

        for element_name in dir(pyOpt):

            element = eval("pyOpt.%s" % element_name)

            try:

                is_subclass = issubclass(element, pyOpt.Optimizer) and not element == pyOpt.Optimizer

            except TypeError:

                continue

            else:

                if is_subclass:

                    optimizers[element_name] = element

        return optimizers

    _pyopt_algorithms = get_pyopt_available_algorithms()

    # Remove algorithms that do not work in our cases
    # 'ALPSO', 'SLSQP', 'SOLVOPT' "ALGENCAN" NSGA2 ALHSO FILTERSD

    for alg in ['ALPSO', 'SLSQP', 'SOLVOPT', 'ALGENCAN', 'NSGA2', 'ALHSO', 'FILTERSD']:

        if alg in _pyopt_algorithms:

            _pyopt_algorithms.pop(alg)

    class PyOptMinimizer(Minimizer):

        def __init__(self, function, parameters, ftol=1e-1, verbosity=10):

            super(PyOptMinimizer, self).__init__(function, parameters, ftol, verbosity)

        def _setup(self):

            self.functor = PyOptWrapper(self.function, self.Npar)

            self._opt_problem = pyOpt.Optimization('Minimum of -log(likelihood)', self.functor)

            self._opt_problem.addObj('f')

            # Add constraints for parameters

            for i, par in enumerate(self.parameters.values()):

                if par.min_value is not None and par.max_value is not None:

                    self._opt_problem.addVar(par.name, 'c', value=par.value, lower=par.min_value, upper=par.max_value)

                elif par.min_value is not None and par.max_value is None:

                    # Lower limited
                    self._opt_problem.addVar(par.name, 'c', value=par.value, lower=par.min_value, upper=np.inf)

                elif par.min_value is None and par.max_value is not None:

                    # upper limited
                    self._opt_problem.addVar(par.name, 'c', value=par.value, lower=-np.inf, upper=par.max_value)

                else:

                    # No limits
                    self._opt_problem.addVar(par.name, 'c', value=par.value, lower=-np.inf, upper=np.inf)

        def set_algorithm(self, algorithm_name):

            assert algorithm_name in _pyopt_algorithms, "Optimizer %s is not part of pyOpt" % algorithm_name

            # Create an instance of the provided optimizer

            self._algorithm_name = algorithm_name

            self._optimizer_instance = _pyopt_algorithms[algorithm_name]()

        def minimize(self, compute_covar=True):

            # Run optimization

            fstr, xstr, inform = self._optimizer_instance(self._opt_problem)

            # Transform to numpy.array

            best_fit_values = np.array(xstr)

            # Compute errors with the Hessian

            if compute_covar:

                covariance_matrix = self._compute_covariance_matrix(best_fit_values)

            else:

                covariance_matrix = None

            self._store_fit_results(best_fit_values, fstr, covariance_matrix)

            return best_fit_values, fstr


    _minimizers["PYOPT"] = PyOptMinimizer


if has_ROOT:

    class FuncWrapper(ROOT.TPyMultiGenFunction):

        def __init__(self, function, dimensions):

            ROOT.TPyMultiGenFunction.__init__(self, self)
            self.function = function
            self.dimensions = int(dimensions)

        def NDim(self):
            return self.dimensions

        def DoEval(self, args):

            new_args = map(lambda i:args[i],range(self.dimensions))

            return self.function(*new_args)


    class ROOTMinimizer(Minimizer):

        def __init__(self, function, parameters, ftol=1e3, verbosity=1):

            super(ROOTMinimizer, self).__init__(function, parameters, ftol, verbosity)

        def _setup(self):

            # Setup the minimizer algorithm

            self.functor = FuncWrapper(self.function, self.Npar)
            self.minimizer = ROOT.Math.Factory.CreateMinimizer("Minuit", "Minimize")
            self.minimizer.Clear()
            self.minimizer.SetMaxFunctionCalls(1000)
            self.minimizer.SetTolerance(0.1)
            self.minimizer.SetPrintLevel(self.verbosity)
            # self.minimizer.SetStrategy(0)

            self.minimizer.SetFunction(self.functor)

            for i, par in enumerate(self.parameters.values()):

                if par.min_value is not None and par.max_value is not None:

                    self.minimizer.SetLimitedVariable(i, par.name, par.value,
                                                      par.delta, par.min_value,
                                                      par.max_value)

                elif par.min_value is not None and par.max_value is None:

                    # Lower limited
                    self.minimizer.SetLowerLimitedVariable(i, par.name, par.value,
                                                           par.delta, par.min_value)

                elif par.min_value is None and par.max_value is not None:

                    # upper limited
                    self.minimizer.SetUpperLimitedVariable(i, par.name, par.value,
                                                           par.delta, par.max_value)

                else:

                    # No limits
                    self.minimizer.SetVariable(i, par.name, par.value, par.delta)

        def minimize(self, compute_covar=True):

            self.minimizer.SetPrintLevel(int(self.verbosity))

            self.minimizer.Minimize()

            best_fit_values = np.array(map(lambda x: x[0], zip(self.minimizer.X(), range(self.Npar))))

            if compute_covar:

                covariance_matrix = self._compute_covariance_matrix(best_fit_values)

            else:

                covariance_matrix = None

            minimum = self.functor(best_fit_values)

            self._store_fit_results(best_fit_values, minimum, covariance_matrix)

            return best_fit_values, minimum

    _minimizers["ROOT"] = ROOTMinimizer


class ContourWorker(object):

    def __init__(self, function, minuit_values, minuit_args, minuit_param_1, minuit_param_2, name_to_position):

        self._minuit_values = minuit_values

        # Update the values for the parameters with the best fit one

        for key, value in self._minuit_values.iteritems():
            minuit_args[key] = value

        # This is a likelihood
        minuit_args['errordef'] = 0.5

        # Disable printing by iminuit

        minuit_args['print_level'] = 0

        self._minuit_args = minuit_args

        # Store the name of the parameters

        self.minuit_param_1 = minuit_param_1
        self.minuit_param_2 = minuit_param_2

        # Store the function
        self._function = function

        # This is a dictionary which gives the ordinal place for a given parameter.
        # It is used in the corner case where the function has only two parameters,
        # to figure out which is the correct order

        self.name_to_position = name_to_position

    def _create_new_minuit_object(self, args):

        # Now create the new minimizer

        _contour_minuit = Minuit(self._function, **args)

        _contour_minuit.tol = 100

        return _contour_minuit

    def __call__(self, args):

        # Get the values for the parameters
        # If we are stepping in only one direction, value_2 will be nan

        value_1, value_2 = args

        # NOTE: unfortunately iminuit does not allow to change the value of a fixed parameter after
        # the creation of the Minuit class. Hence we need to create a new class each time,
        # which sucks

        # Create a copy of the init args for Minuit

        this_minuit_args = dict(self._minuit_args)

        # Now set the parameters under scrutiny to the current values

        this_minuit_args[self.minuit_param_1] = value_1

        if self.minuit_param_2 is not None:
            this_minuit_args[self.minuit_param_2] = value_2

        # Fix the parameters under scrutiny

        for minuit_name in [self.minuit_param_1, self.minuit_param_2]:

            if minuit_name is None:
                # Only one parameter to analyze

                continue

            if minuit_name not in this_minuit_args.keys():

                raise ParameterIsNotFree("Parameter %s is not a free parameter." % minuit_name)

            else:

                this_minuit_args['fix_%s' % minuit_name] = True

        # Finally create a new minimizer
        this_contour_minuit = self._create_new_minuit_object(this_minuit_args)

        # Handle the corner case where there are no free parameters
        # after fixing the two under scrutiny

        if len(this_contour_minuit.list_of_vary_param()) == 0:

            # All parameters are fixed, just return the likelihood function

            if self.minuit_param_2 is None:

                value = self._function(value_1)

            else:

                # This is needed because the user could specify the
                # variables in a different order than what is specified in the calling sequence
                # of f

                this_variables = [0, 0]
                this_variables[self.name_to_position[self.minuit_param_1]] = value_1
                this_variables[self.name_to_position[self.minuit_param_2]] = value_2

                value = self._function(*this_variables)

            return value

        try:

            this_contour_minuit.migrad()

        # In the following except I cannot catch specific exceptions because I don't exactly know which kind
        # of exception migrad can raise...

        except:

            # In this context this is not such a big deal,
            # because we might be so far from the minimum that
            # the fit cannot converge

            return FIT_FAILED

        return this_contour_minuit.fval



__author__ = 'giacomov'

from GenericOGIPLike import GenericOGIPLike

__instrument_name = "VERITAS"

class VERITASLike(GenericOGIPLike):

    def __init__(self, name, udara_style_root_file):

        super(VERITASLike, self).__init__(name, )


"""
Define the interface for a plugin class.
"""

import abc


class PluginPrototype(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def set_model(self, likelihood_model_instance):
        """
        Set the model to be used in the joint minimization. Must be a LikelihoodModel instance.
        """
        pass

    @abc.abstractmethod
    def get_name(self):
        """
        Return a name for this data set (likely set during the constructor)
        """
        pass

    @abc.abstractmethod
    def get_log_like(self):
        """
        Return the value of the log-likelihood with the current values for the
        parameters
        """
        pass

    @abc.abstractmethod
    def get_nuisance_parameters(self):
        """
        Return a list of nuisance parameters. Return an empty list if there
        are no nuisance parameters
        """
        pass

    @abc.abstractmethod
    def inner_fit(self):
        """
        This is used for the profile likelihood. Keeping fixed all parameters in the
        LikelihoodModel, this method minimize the logLike over the remaining nuisance
        parameters, i.e., the parameters belonging only to the model for this
        particular detector. If there are no nuisance parameters, simply return the
        logLike value.
        """
        pass



# This module handle the lazy dependence on IPython


def fallback_display(x):

    print(x)

try:

    from IPython.core.display import display

except ImportError:

    display = fallback_display





import collections
import time
import numpy
import scipy.optimize
import scipy.stats
import matplotlib.pyplot as plt
import matplotlib.cm as cm
from matplotlib.colors import BoundaryNorm

from threeML.io.rich_display import display
import numpy as np
import pandas as pd
import uncertainties


from threeML.minimizer import minimization
from threeML.exceptions import custom_exceptions
from threeML.io.table import Table, NumericMatrix
from threeML.parallel.parallel_client import ParallelClient
from threeML.io.progress_bar import ProgressBar
from threeML.config.config import threeML_config
from threeML.exceptions.custom_exceptions import custom_warnings, FitFailed
from threeML.utils.uncertainties_regexpr import get_uncertainty_tokens


from astromodels import ModelAssertionViolation

class ReducingNumberOfThreads(Warning):
    pass


class ReducingNumberOfSteps(Warning):
    pass


class NotANumberInLikelihood(Warning):
    pass


class JointLikelihood(object):
    def __init__(self, likelihood_model, data_list, verbose=False):
        """
        Implement a joint likelihood analysis.

        :param likelihood_model: the model for the likelihood analysis
        :param data_list: the list of data sets (plugin instances) to be used in this analysis
        :param verbose: (True or False) print every step in the -log likelihood minimization
        :return:
        """

        # Process optional keyword parameters
        self.verbose = verbose

        self._likelihood_model = likelihood_model

        self._data_list = data_list

        for dataset in self._data_list.values():

            dataset.set_model(self._likelihood_model)

        # This is to keep track of the number of calls to the likelihood
        # function
        self.ncalls = 0

        # Pre-defined minimizer is Minuit

        self.set_minimizer("MINUIT")

        # Initial set of free parameters

        self._free_parameters = self._likelihood_model.free_parameters

        # Initially set the value of _current_minimum to None, it will be change by the fit() method

        self._current_minimum = None

        self._minimizer = None

    @property
    def likelihood_model(self):
        """
        :return: likelihood model for this analysis
        """

        return self._likelihood_model

    @property
    def data_list(self):
        """
        :return: data list for this analysis
        """

        return self._data_list

    @property
    def current_minimum(self):
        """
        :return: current minimum of the joint likelihood (available only after the fit() method)
        """
        return self._current_minimum

    @property
    def minimizer(self):
        """
        :return: an instance of the minimizer used in the fit (available only after the fit() method)
        """
        return self._minimizer

    @property
    def covariance_matrix(self):
        """
        :return: covariance matrix from the last fit
        """
        try:

            return self._minimizer.covariance_matrix

        except AttributeError:

            raise RuntimeError("You need to run a fit before accessing the covariance matrix")

    @property
    def correlation_matrix(self):
        """
        :return: correlation matrix from the last fit
        """

        try:

            return self._minimizer.correlation_matrix

        except AttributeError:

            raise RuntimeError("You need to run a fit before accessing the correlation matrix")

    def _update_free_parameters(self):

        """Update the dictionary of free parameters"""

        self._free_parameters = self._likelihood_model.free_parameters

    def fit(self):
        """
        Perform a fit of the current likelihood model on the datasets

        :param pre_fit: (True or False) If True, perform a pre-fit with only normalizations free (experimental)
        :return: a dictionary with the results on the parameters, and the values of the likelihood at the minimum
                 for each dataset and the total one.
        """

        # Update the list of free parameters, to be safe against changes the user might do between
        # the creation of this class and the calling of this method

        self._update_free_parameters()

        # Now check and fix if needed all the deltas of the parameters
        # to 20% of their value (otherwise the fit will be super-slow)

        for k, v in self._free_parameters.iteritems():

            if abs(v.delta) < abs(v.value) * 0.2:

                v.delta = abs(v.value) * 0.2

        # Instance the minimizer

        self._minimizer = self._get_minimizer(self.minus_log_like_profile,
                                              self._free_parameters)

        # Perform the fit

        xs, log_likelihood_minimum = self._minimizer.minimize()

        if log_likelihood_minimum == minimization.FIT_FAILED:

            raise FitFailed("The fit failed to converge.")

        # Store the current minimum for the -log likelihood

        self._current_minimum = float(log_likelihood_minimum)

        # Print results

        # Get the results from the minimizer (a panda container)

        fit_results = self._minimizer.fit_results

        # Now produce an ad-hoc display. We don't use the pandas display methods because
        # we want to display uncertainties with the right number of significant numbers

        data = []

        # Also store the maximum length to decide the length for the line

        name_length = 0

        for i, parameter_name in enumerate(fit_results.index.values):

            value = fit_results.at[parameter_name, 'value']

            error = fit_results.at[parameter_name, 'error']

            # Format the value and the error with sensible significant
            # numbers
            x = uncertainties.ufloat(value, error)

            # Add some space around the +/- sign

            rep = x.__str__().replace("+/-", " +/- ")

            data.append([i, parameter_name, rep, self._free_parameters[parameter_name].unit])

            if len(parameter_name) > name_length:

                name_length = len(parameter_name)

        best_fit_table = Table(rows=data,
                      names=["#", "Name", "Best fit value", "Unit"],
                      dtype=(str, 'S%i' % name_length, str, str))

        print("Best fit values:\n")

        display(best_fit_table)

        print("\nNOTE: errors on parameters are approximate. Use get_errors().\n")

        # Now display the nuisance parameters

        nuisance_parameters = collections.OrderedDict()

        for dataset in self._data_list.values():

            for pName in dataset.get_nuisance_parameters():

                nuisance_parameters[(dataset.get_name(), pName)] = dataset.nuisanceParameters[pName]

        nuisance_parameters_table = self._get_table_of_parameters(nuisance_parameters)

        print("Nuisance parameters:\n")

        display(nuisance_parameters_table)

        print("\nCorrelation matrix:\n")

        best_fit_table = NumericMatrix(self._minimizer.correlation_matrix)

        for col in best_fit_table.colnames:

            best_fit_table[col].format = '2.2f'

        display(best_fit_table)

        # Now collect the values for the likelihood for the various datasets

        # First restore best fit (to make sure we compute the likelihood at the right point)
        self._minimizer.restore_best_fit()

        # Fill the dictionary with the values of the -log likelihood (total, and dataset by dataset)

        minus_log_likelihood_values = collections.OrderedDict()

        minus_log_likelihood_values['all'] = self._current_minimum

        total = 0

        for dataset in self._data_list.values():

            ml = dataset.inner_fit() * (-1)

            minus_log_likelihood_values[dataset.get_name()] = ml

            total += ml

        assert total == self._current_minimum, "Current minimum stored after fit and current do not correspond!"

        print("\nValues of -log(likelihood) at the minimum:\n")

        loglike_dataframe = pd.DataFrame(minus_log_likelihood_values.items(), columns=['instrument','-log(likelihood)'])

        display(loglike_dataframe)

        return fit_results, loglike_dataframe

    def get_errors(self):
        """
        Compute the errors on the parameters using the profile likelihood method.

        :return: a dictionary containing the asymmetric errors for each parameter.
        """

        # Check that the user performed a fit first

        assert self._current_minimum is not None, "You have to run the .fit method before calling errors."

        errors = self._minimizer.get_errors()

        # Set the parameters back to the best fit value
        self.restore_best_fit()

        # Print a table with the errors

        data = []
        name_length = 0

        for k, v in self._free_parameters.iteritems():

            # Format the value and the error with sensible significant
            # numbers

            # Process the negative error

            neg_error = abs(errors[k][0])

            if np.isnan(neg_error):

                x = uncertainties.ufloat(v.value, 0)

            else:

                x = uncertainties.ufloat(v.value, neg_error)

            # Split the uncertainty in number, negative error, and exponent (if any)

            num, uncm, exponent = get_uncertainty_tokens(x)

            if np.isnan(neg_error):

                uncm = np.nan

            # Process the positive error

            pos_error = abs(errors[k][1])

            if np.isnan(pos_error):

                x = uncertainties.ufloat(v.value, 0)

            else:

                x = uncertainties.ufloat(v.value, pos_error)

            # Split the uncertainty in number, positive error, and exponent (if any)

            _, uncp, _ = get_uncertainty_tokens(x)

            if np.isnan(pos_error):

                uncp = np.nan

            if exponent is None:

                # Number without exponent

                pretty_string = "%s -%s +%s" % (num, uncm, uncp)

            else:

                # Number with exponent

                pretty_string = "(%s -%s +%s)%s" % (num, uncm, uncp, exponent)

            data.append([k, pretty_string, v.unit])

            if len(k) > name_length:
                name_length = len(k)

        # Create and display the table

        table = Table(rows=data,
                      names=["Name", "Value", "Unit"],
                      dtype=('S%i' % name_length, str, str))

        display(table)

    def get_contours(self, param_1, param_1_minimum, param_1_maximum, param_1_n_steps,
                     param_2=None, param_2_minimum=None, param_2_maximum=None, param_2_n_steps=None,
                     progress=True, **options):
        """
        Generate confidence contours for the given parameters by stepping for the given number of steps between
        the given boundaries. Call it specifying only source_1, param_1, param_1_minimum and param_1_maximum to
        generate the profile of the likelihood for parameter 1. Specify all parameters to obtain instead a 2d
        contour of param_1 vs param_2.

        NOTE: if using parallel computation, param_1_n_steps must be an integer multiple of the number of running
        engines. If that is not the case, the code will reduce the number of steps to match that requirement, and
        issue a warning

        :param param_1: fully qualified name of the first parameter or parameter instance
        :param param_1_minimum: lower bound for the range for the first parameter
        :param param_1_maximum: upper bound for the range for the first parameter
        :param param_1_n_steps: number of steps for the first parameter
        :param param_2: fully qualified name of the second parameter or parameter instance
        :param param_2_minimum: lower bound for the range for the second parameter
        :param param_2_maximum: upper bound for the range for the second parameter
        :param param_2_n_steps: number of steps for the second parameter
        :param progress: (True or False) whether to display progress or not
        :param log: by default the steps are taken linearly. With this optional parameter you can provide a tuple of
                    booleans which specify whether the steps are to be taken logarithmically. For example,
                    'log=(True,False)' specify that the steps for the first parameter are to be taken logarithmically,
                    while they are linear for the second parameter. If you are generating the profile for only one
                    parameter, you can specify 'log=(True,)' or 'log=(False,)' (optional)
        :return: a tuple containing an array corresponding to the steps for the first parameter, an array corresponding
                 to the steps for the second parameter (or None if stepping only in one direction), a matrix of size
                 param_1_steps x param_2_steps containing the value of the function at the corresponding points in the
                 grid. If param_2_steps is None (only one parameter), then this reduces to an array of
                 size param_1_steps.
        """

        if hasattr(param_1,"value"):

            # Substitute with the name
            param_1 = param_1.path

        if hasattr(param_2,'value'):

            param_2 = param_2.path

        # Check that the parameters exist
        assert param_1 in self._likelihood_model.free_parameters, "Parameter %s is not a free parameters of the " \
                                                                 "current model" % param_1

        if param_2 is not None:
            assert param_2 in self._likelihood_model.free_parameters, "Parameter %s is not a free parameters of the " \
                                                                      "current model" % param_2


        # Check that we have a valid fit

        assert self._current_minimum is not None, "You have to run the .fit method before calling get_contours."

        # Then restore the best fit

        self._minimizer.restore_best_fit()

        # Check minimal assumptions about the procedure

        assert not (param_1 == param_2), "You have to specify two different parameters"

        assert param_1_minimum < param_1_maximum, "Minimum larger than maximum for parameter 1"

        if param_2 is not None:
            assert param_2_minimum < param_2_maximum, "Minimum larger than maximum for parameter 2"

        # Check whether we are parallelizing or not

        if not threeML_config['parallel']['use-parallel']:

            a, b, cc = self.minimizer.contours(param_1, param_1_minimum, param_1_maximum, param_1_n_steps,
                                               param_2, param_2_minimum, param_2_maximum, param_2_n_steps,
                                               progress, **options)

            # Collapse the second dimension of the results if we are doing a 1d contour

            if param_2 is None:
                cc = cc[:, 0]

        else:

            # With parallel computation

            # In order to distribute fairly the computation, the strategy is to parallelize the computation
            # by assigning to the engines one "line" of the grid at the time

            # Connect to the engines

            client = ParallelClient(**options)

            # Get the number of engines

            n_engines = client.get_number_of_engines()

            # Check whether the number of threads is larger than the number of steps in the first direction

            if n_engines > param_1_n_steps:

                n_engines = int(param_1_n_steps)

                custom_warnings.warn("The number of engines is larger than the number of steps. Using only %s engines."
                                     % n_engines, ReducingNumberOfThreads)

            # Check if the number of steps is divisible by the number
            # of threads, otherwise issue a warning and make it so

            if float(param_1_n_steps) % n_engines != 0:
                # Set the number of steps to an integer multiple of the engines
                # (note that // is the floor division, also called integer division)

                param_1_n_steps = (param_1_n_steps // n_engines) * n_engines

                custom_warnings.warn("Number of steps is not a multiple of the number of threads. Reducing steps to %s"
                                     % param_1_n_steps, ReducingNumberOfSteps)

            # Compute the number of splits, i.e., how many lines in the grid for each engine.
            # (note that this is guaranteed to be an integer number after the previous checks)

            p1_split_steps = param_1_n_steps // n_engines

            # Prepare arrays for results

            if param_2 is None:

                # One array
                pcc = numpy.zeros(param_1_n_steps)

                pa = numpy.linspace(param_1_minimum, param_1_maximum, param_1_n_steps)
                pb = None

            else:

                pcc = numpy.zeros((param_1_n_steps, param_2_n_steps))

                # Prepare the two axes of the parameter space
                pa = numpy.linspace(param_1_minimum, param_1_maximum, param_1_n_steps)
                pb = numpy.linspace(param_2_minimum, param_2_maximum, param_2_n_steps)

            # Define the parallel worker which will go through the computation

            # NOTE: I only divide
            # on the first parameter axis so that the different
            # threads are more or less well mixed for points close and
            # far from the best fit

            def worker(start_index):

                # Re-create the minimizer

                backup_freeParameters = map(lambda x:x.value, self._likelihood_model.free_parameters.values())

                this_minimizer = self._get_minimizer(self.minus_log_like_profile,
                                                     self._free_parameters)

                this_p1min = pa[start_index * p1_split_steps]
                this_p1max = pa[(start_index + 1) * p1_split_steps - 1]

                # print("From %s to %s" % (this_p1min, this_p1max))

                aa, bb, ccc = this_minimizer.contours(param_1, this_p1min, this_p1max, p1_split_steps,
                                                      param_2, param_2_minimum, param_2_maximum,
                                                      param_2_n_steps,
                                                      False, **options)

                # Restore best fit values

                for val, par in zip(backup_freeParameters, self._likelihood_model.free_parameters.values()):

                    par.value = val

                return ccc

            # Get a balanced view of the engines

            lview = client.load_balanced_view()
            # lview.block = True

            # Distribute the work among the engines and start it, but return immediately the control
            # to the main thread

            amr = lview.map_async(worker, range(n_engines))

            # print progress

            progress = ProgressBar(n_engines)

            # This loop will check from time to time the status of the computation, which is happening on
            # different threads, and update the progress bar

            while not amr.ready():
                # Check and report the status of the computation every second

                time.sleep(1 + np.random.uniform(0, 1))

                # if (debug):
                #     stdouts = amr.stdout
                #
                #     # clear_output doesn't do much in terminal environments
                #     for stdout, stderr in zip(amr.stdout, amr.stderr):
                #         if stdout:
                #             print "%s" % (stdout[-1000:])
                #         if stderr:
                #             print "%s" % (stderr[-1000:])
                #     sys.stdout.flush()

                progress.animate(amr.progress - 1)

            # Always display 100% at the end

            progress.animate(n_engines - 1)

            # Add a new line after the progress bar
            print("\n")

            # print("Serial time: %1.f (speed-up: %.1f)" %(amr.serial_time, float(amr.serial_time) / amr.wall_time))

            # Get the results. This will raise exceptions if something wrong happened during the computation.
            # We don't catch it so that the user will be aware of that

            res = amr.get()

            # Now re-assemble the vector of results taking the different parts from the engines

            for i in range(n_engines):

                if param_2 is None:

                    pcc[i * p1_split_steps: (i + 1) * p1_split_steps] = res[i][:, 0]

                else:

                    pcc[i * p1_split_steps: (i + 1) * p1_split_steps, :] = res[i]

            # Give the results the names that the following code expect. These are kept separate for debugging
            # purposes

            cc = pcc
            a = pa
            b = pb

        # Here we have done the computation, in parallel computation or not. Let's make the plot
        # with the contour

        if param_2 is not None:

            # 2d contour

            fig = self._plot_contours("%s" % (param_1), a, "%s" % (param_2,), b, cc)

        else:

            # 1d contour (i.e., a profile)

            fig = self._plot_profile("%s" % (param_1), a, cc)

        # Check if we found a better minimum. This shouldn't happen, but in case of very difficult fit
        # it might.

        if self._current_minimum - cc.min() > 0.1:

            if param_2 is not None:

                idx = cc.argmin()

                aidx, bidx = numpy.unravel_index(idx, cc.shape)

                print("\nFound a better minimum: %s with %s = %s and %s = %s. Run again your fit starting from here."
                      % (cc.min(), param_1, a[aidx], param_2, b[bidx]))

            else:

                idx = cc.argmin()

                print("Found a better minimum: %s with %s = %s. Run again your fit starting from here."
                      % (cc.min(), param_1, a[idx]))

        return a, b, cc, fig

    def minus_log_like_profile(self, *trial_values):
        """
        Return the minus log likelihood for a given set of trial values

        :param trial_values: the trial values. Must be in the same number as the free parameters in the model
        :return: minus log likelihood
        """

        # Keep track of the number of calls
        self.ncalls += 1

        # Transform the trial values in a numpy array

        trial_values = numpy.array(trial_values)

        # Check that there are no nans within the trial values

        # This is the fastest way to check for any nan
        # (try other methods if you don't believe me)

        if not numpy.isfinite(numpy.dot(trial_values, trial_values.T)):
            # There are nans, something weird is going on. Return FIT_FAILED so the engine
            # stays away from this (or fail)

            return minimization.FIT_FAILED

        # Assign the new values to the parameters

        for i, parameter in enumerate(self._free_parameters.values()):

            parameter.value = trial_values[i]

        # Now profile out nuisance parameters and compute the new value
        # for the likelihood

        summed_log_likelihood = 0

        for dataset in self._data_list.values():

            try:

                this_log_like = dataset.inner_fit()

            except ModelAssertionViolation:

                # This is a zone of the parameter space which is not allowed. Return
                # a big number for the likelihood so that the fit engine will avoid it

                custom_warnings.warn("Fitting engine in forbidden space: %s" % (trial_values,),
                                     custom_exceptions.ForbiddenRegionOfParameterSpace)

                return minimization.FIT_FAILED

            except:

                # Do not intercept other errors

                raise

            # if this_log_like > 0:

            #    raise RuntimeError("LogLike cannot be larger than 0")

            summed_log_likelihood += this_log_like

            # dataset.getLogLike()

        # Check that the global like is not NaN
        # I use this weird check because it is not guaranteed that the plugins return np.nan,
        # especially if they are written in something other than python

        if "%s" % summed_log_likelihood == 'nan':
            custom_warnings.warn("These parameters returned a logLike = Nan: %s" % (trial_values,),
                                 NotANumberInLikelihood)

            return minimization.FIT_FAILED

        if self.verbose:
            print("Trying with parameters %s, resulting in logL = %s" % (trial_values, summed_log_likelihood))

        # Return the minus log likelihood

        return summed_log_likelihood * (-1)

    def set_minimizer(self, minimizer, algorithm=None):
        """
        Set the minimizer to be used, among those available. At the moment these are supported:

        * ROOT
        * MINUIT (which means iminuit, default)
        * PYOPT

        :param minimizer: the name of the new minimizer.
        :param algorithm: (optional) for the PYOPT minimizer, specify the algorithm among those available. See a list at
        http://www.pyopt.org/reference/optimizers.html . For the other minimizers this is ignored, if provided.
        :return: (none)
        """

        if minimizer.upper() == "MINUIT":

            self._minimizer_name = "MINUIT"
            self._minimizer_algorithm = None

        elif minimizer.upper() == "ROOT":

            assert minimization.has_ROOT, "pyROOT not available on this system. Cannot use ROOT minimizer."

            self._minimizer_name = "ROOT"
            self._minimizer_algorithm = None

        elif minimizer.upper() == "PYOPT":

            assert minimization.has_pyOpt, "pyOpt not available on this system. Cannot use PYOPT minimizer."

            assert algorithm is not None, "You have to provide an algorithm for the PYOPT minimizer. " \
                                          "See http://www.pyopt.org/reference/optimizers.html"

            self._minimizer_name = "PYOPT"

            assert algorithm.upper() in minimization._pyopt_algorithms, "Provided algorithm not available. " \
                                                                        "Available algorithms are: " \
                                                                        "%s" % minimization._pyopt_algorithms.keys()
            self._minimizer_algorithm = algorithm.upper()

        else:

            raise ValueError("Do not know minimizer %s" % minimizer)

        # Get the type

        self._minimizer_type = minimization.get_minimizer(self._minimizer_name)

    def _get_minimizer(self, *args, **kwargs):

        minimizer_instance = self._minimizer_type(*args, **kwargs)

        if self._minimizer_algorithm:

            minimizer_instance.set_algorithm(self._minimizer_algorithm)

        return minimizer_instance

    def restore_best_fit(self):
        """
        Restore the model to its best fit

        :return: (none)
        """

        if self._minimizer:

            self._minimizer.restore_best_fit()

        else:

            custom_warnings.warn("Cannot restore best fit, since fit has not been executed.")

    def _get_table_of_parameters(self, parameters):

        data = []
        max_length_of_name = 0

        for k, v in parameters.iteritems():

            current_name = "%s_of_%s" % (k[1], k[0])

            data.append([current_name, "%s" % v.value, v.unit])

            if len(v.name) > max_length_of_name:
                max_length_of_name = len(current_name)

        table = Table(rows=data,
                      names=["Name", "Value", "Unit"],
                      dtype=('S%i' % max_length_of_name, str, 'S15'))

        return table

    def _plot_profile(self, name1, a, cc):
        """
        Plot the likelihood profile.

        :param name1: Name of parameter
        :param a: grid for the parameter
        :param cc: log. likelihood values for the parameter
        :return: a figure containing the likelihood profile
        """

        # plot 1,2 and 3 sigma horizontal lines

        sigmas = [1, 2, 3]

        # Compute the corresponding probability. We do not
        # pre-compute them because we will introduce at
        # some point the possibility to personalize the
        # levels

        probabilities = []

        for s in sigmas:
            # One-sided probability
            # It is one-sided because we consider one side at the time
            # when computing the error

            probabilities.append(1 - (scipy.stats.norm.sf(s) * 2))

        # Compute the corresponding delta chisq. (chisq has 1 d.o.f.)

        # noinspection PyTypeChecker
        delta_chi2 = np.array(scipy.stats.chi2.ppf(probabilities, 1) / 2.0)  # two-sided!

        fig = plt.figure()
        sub = fig.add_subplot(111)

        # Neutralize values of the loglike too high
        # (fit failed)
        idx = (cc == minimization.FIT_FAILED)

        sub.plot(a[~idx], cc[~idx], lw=2)

        # Now plot the failed fits as "x"

        sub.plot(a[idx], [cc.min()] * a[idx].shape[0], 'x', c='red', markersize=2)

        # Decide colors
        colors = ['blue', 'cyan', 'red']

        for s, d, c in zip(sigmas, delta_chi2, colors):
            sub.axhline(self._current_minimum + d, linestyle='--',
                        color=c, label=r'%s $\sigma$' % s, lw=2)

        # Fix the axis to cover from the minimum to the 3 sigma line
        sub.set_ylim([self._current_minimum - delta_chi2[0],
                      self._current_minimum + (delta_chi2[-1] * 2)])

        plt.legend(loc=0, frameon=True)

        sub.set_xlabel(name1)
        sub.set_ylabel("-log( likelihood )")

        return fig

    def _plot_contours(self, name1, a, name2, b, cc):
        """
        Make a contour plot.

        :param name1: Name of the first parameter
        :param a: Grid for the first parameter (dimension N)
        :param name2: Name of the second parameter
        :param b: grid for the second parameter (dimension M)
        :param cc: N x M matrix containing the value of the log.likelihood for each point in the grid
        :return: figure containing the contour
        """

        # Check that we have something to plot

        delta = cc.max() - cc.min()

        if delta < 0.5:
            print("\n\nThe maximum difference in statistic is %s among all the points in the grid." % delta)
            print(" This is too small. Enlarge the search region to display a contour plot")

            return None

        # plot 1,2 and 3 sigma contours
        sigmas = [1, 2, 3]

        # Compute the corresponding probability. We do not
        # pre-compute them because we will introduce at
        # some point the possibility to personalize the
        # levels
        probabilities = []

        for s in sigmas:
            # One-sided probability
            # It is one-sided because we consider one side at the time
            # when computing the error

            probabilities.append(1 - (scipy.stats.norm.sf(s) * 2))

        # Compute the corresponding delta chisq. (chisq has 2 d.o.f.)
        delta_chi2 = scipy.stats.chi2.ppf(probabilities, 2) / 2.0  # two-sided

        # Boundaries for the colormap
        bounds = [self._current_minimum]
        # noinspection PyTypeChecker
        bounds.extend(self._current_minimum + delta_chi2)
        bounds.append(cc.max())

        # Define the color palette
        palette = cm.Pastel1
        palette.set_over('white')
        palette.set_under('white')
        palette.set_bad('white')

        fig = plt.figure()
        sub = fig.add_subplot(111)

        # Show the contours with square axis

        # NOTE: suppress the UnicodeWarning, which is due to a small problem in matplotlib

        with custom_warnings.catch_warnings():

            # Cause all warnings to always be triggered.
            custom_warnings.simplefilter("ignore", UnicodeWarning)

            im = sub.imshow(cc,
                            cmap=palette,
                            extent=[b.min(), b.max(), a.min(), a.max()],
                            aspect=float(b.max() - b.min()) / (a.max() - a.min()),
                            origin='lower',
                            norm=BoundaryNorm(bounds, 256),
                            interpolation='bicubic',
                            vmax=(self._current_minimum + delta_chi2).max())

        # Plot the color bar with the sigmas
        cb = fig.colorbar(im, boundaries=bounds[:-1])
        lbounds = [0]
        lbounds.extend(bounds[:-1])
        cb.set_ticks(lbounds)
        ll = ['']
        ll.extend(map(lambda x: r'%i $\sigma$' % x, sigmas))
        cb.set_ticklabels(ll)

        # Align the labels to the end of the color level
        for t in cb.ax.get_yticklabels():
            t.set_verticalalignment('baseline')

        # Draw the line contours
        sub.contour(b, a, cc, self._current_minimum + delta_chi2)

        # Set the axes labels

        sub.set_xlabel(name2)
        sub.set_ylabel(name1)

        return fig



import os


def file_existing_and_readable(filename):

    sanitized_filename = sanitize_filename(filename)

    if os.path.exists(sanitized_filename):

        # Try to open it

        try:

            with open(sanitized_filename):

                pass

        except:

            return False

        else:

            return True

    else:

        return False


def sanitize_filename(filename):
    return os.path.expandvars(os.path.expanduser(filename))



__author__ = 'drjfunk'
import math
import re
import warnings

import astropy.io.fits as pyfits
import numpy as np
import scipy.integrate

from threeML.plugins.ogip import gbm_light_curve_plot
from threeML.plugin_prototype import PluginPrototype
from threeML.plugins.ogip import OGIPPluginPGstat

__instrument_name = "Fermi GBM TTE (all detectors)"


class FermiGBMLikeTTE(OGIPPluginPGstat, PluginPrototype):

    def __init__(self, name, ttefile, bkgselections, srcinterval, rspfile, polyorder=-1):
        """
        If the input files are TTE files. Background selections are specified as
        a comma separated string e.g. "-10-0,10-20"

        Initial source selection is input as a string e.g. "0-5"

        One can choose a background polynomial order by hand (up to 4th order)
        or leave it as the default polyorder=-1 to decide by LRT test
        
        FermiGBM_TTE_Like("GBM","glg_tte_n6_bn080916412.fit","-10-0,10-20","0-5","rspfile.rsp{2}")
        to load the second spectrum, second background spectrum and second response.
        """

        self.name = name

        OGIPPluginPGstat.__init__(self, name, rspfile, ttefile=ttefile)

        self._polyorder = polyorder

        self.ttefile = GBMTTEFile(ttefile)

        self._backgroundexists = False
        self._energyselectionexists = False

        # Start with an empty mask (the user will overwrite it using the
        # setActiveMeasurement method)
        self.mask = np.asarray(np.ones(self.ttefile.nchans), np.bool)

        # Fit the background and
        # Obtain the counts for the initial input interval
        # which is embeded in the background call

        # First get the initial tmin and tmax

        self.tmin, self.tmax = self._parse_time_interval(srcinterval)

        self.set_background_interval(*bkgselections.split(','))

    @staticmethod
    def _parse_time_interval(time_interval):

        # The following regular expression matches any two numbers, positive or negative,
        # like "-10 --5","-10 - -5", "-10-5", "5-10" and so on

        tokens = re.match('(\-?\+?[0-9]+\.?[0-9]*)\s*-\s*(\-?\+?[0-9]+\.?[0-9]*)', time_interval).groups()

        return map(float, tokens)

    def set_active_time_interval(self, arg):
        '''Set the time interval to be used during the analysis.
        For now, only one interval can be selected. This may be
        updated in the future to allow for self consistent time
        resolved analysis.
        Specified as 'tmin-tmax'. Intervals are in seconds. Example:

        set_active_time_interval("0.0-10.0")

        which will set the energy range 0-10. seconds.
        '''

        self.tmin, self.tmax = self._parse_time_interval(arg)





        # First build the mas for the events in time
        timemask = np.logical_and(self.ttefile.events - self.ttefile.triggertime >= self.tmin,
                                  self.ttefile.events - self.ttefile.triggertime <= self.tmax)
        tmpcounts = []  # Temporary list to hold the total counts per chan
        tmpbackground = []  # Temporary list to hold the background counts per chan
        tmpbackgrounderr = []  # Temporary list to hold the background counts per chan

        for chan in range(self.ttefile.nchans):
            channelmask = self.ttefile.pha == chan
            countsmask = np.logical_and(channelmask, timemask)
            totalcounts = len(self.ttefile.events[countsmask])

            # Now integrate the appropriate background polynomial
            backgroundcounts = self._polynomials[chan].integral(self.tmin, self.tmax)
            backgrounderror = self._polynomials[chan].integralError(self.tmin, self.tmax)

            tmpcounts.append(totalcounts)
            tmpbackground.append(backgroundcounts)
            tmpbackgrounderr.append(backgrounderror)

        counts = np.array(tmpcounts)
        bkgCounts = np.array(tmpbackground)
        bkgErr = np.array(tmpbackgrounderr)

        # Calculate the exposure using the GBM dead time (Meegan et al. 2009)
        totaldeadtime = self.ttefile.deadtime[timemask].sum()

        exposure = (self.tmax - self.tmin) - totaldeadtime

        # Run through the proper checks on counts
        self._initialSetup(self.mask, counts, bkgCounts, exposure, bkgErr)

    def set_background_interval(self, *time_intervals_spec):
        '''Set the time interval to fit the background.
        Multiple intervals can be input as separate arguments
        Specified as 'tmin-tmax'. Intervals are in seconds. Example:

        setBackgroundInterval("-10.0-0.0","10.-15.")
        '''

        self._backgroundselections = []

        for time_interval_spec in time_intervals_spec:

            t1, t2 = self._parse_time_interval(time_interval_spec)

            self._backgroundselections.append((t1,t2))

        self._backgroundselections = np.array(self._backgroundselections)

        # Fit the background with the given intervals
        self._FitBackground()

        # Since changing the background will alter the counts
        # We need to recalculate the source interval
        self.set_active_time_interval("%.5f-%.5f" % (self.tmin, self.tmax))

    def view_lightcurve(self, start=-10, stop=20., dt=1.):

        binner = np.arange(start, stop + dt, dt)
        cnts, bins = np.histogram(self.ttefile.events - self.ttefile.triggertime, bins=binner)
        time_bins = np.array([[bins[i], bins[i + 1]] for i in range(len(bins) - 1)])

        bkg = []
        for tb in time_bins:
            tmpbkg = 0.  # Maybe I can do this perenergy at some point
            for poly in self._polynomials:
                tmpbkg += poly.integral(tb[0], tb[1]) / (dt)

            bkg.append(tmpbkg)

        gbm_light_curve_plot(time_bins, cnts, bkg, dt, selection=[self.tmin, self.tmax])

    def _fitGlobalAndDetermineOptimumGrade(self, cnts, bins):
        # Fit the sum of all the channels to determine the optimal polynomial
        # grade
        Nintervals = len(bins)

        # y                         = []
        # for i in range(Nintervals):
        #  y.append(np.sum(counts[i]))
        # pass
        # y                         = np.array(y)

        # exposure                  = np.array(data.field("EXPOSURE"))

        print("\nLooking for optimal polynomial grade:")

        # Fit all the polynomials

        minGrade = 0
        maxGrade = 4
        logLikelihoods = []

        for grade in range(minGrade, maxGrade + 1):

            polynomial, logLike = self._polyfit(bins, cnts, grade)

            logLikelihoods.append(logLike)

        # Found the best one
        deltaLoglike = np.array(map(lambda x: 2 * (x[0] - x[1]), zip(logLikelihoods[:-1], logLikelihoods[1:])))

        print("\ndelta log-likelihoods:")

        for i in range(maxGrade):

            print("%s -> %s: delta Log-likelihood = %s" % (i, i + 1, deltaLoglike[i]))


        print("")

        deltaThreshold = 9.0

        mask = (deltaLoglike >= deltaThreshold)

        if (len(mask.nonzero()[0]) == 0):

            # best grade is zero!
            bestGrade = 0

        else:

            bestGrade = mask.nonzero()[0][-1] + 1

        return bestGrade

    def _FitBackground(self):

        self._backgroundexists = True
        ## Seperate everything by energy channel

        # Select all the events that are in the background regions
        # and make a mask

        allbkgmasks = []

        for bkgsel in self._backgroundselections:

            allbkgmasks.append(np.logical_and(self.ttefile.events - self.ttefile.triggertime >= bkgsel[0],
                                              self.ttefile.events - self.ttefile.triggertime <= bkgsel[1]))
        backgroundmask = allbkgmasks[0]

        # If there are multiple masks:
        if len(allbkgmasks) > 1:
            for mask in allbkgmasks[1:]:
                backgroundmask = np.logical_or(backgroundmask, mask)

        # Now we will find the the best poly order unless the use specified one
        # The total cnts (over channels) is binned to 1 sec intervals
        if self._polyorder == -1:
            totalbkgevents = self.ttefile.events[backgroundmask]
            binwidth = 1.
            cnts, bins = np.histogram(totalbkgevents - self.ttefile.triggertime,
                                      bins=np.arange(self.ttefile.startevents - self.ttefile.triggertime,
                                                     self.ttefile.stopevents - self.ttefile.triggertime,
                                                     binwidth))

            cnts = cnts/binwidth
            # Find the mean time of the bins
            meantime = []
            for i in xrange(len(bins) - 1):
                m = np.mean((bins[i], bins[i + 1]))
                meantime.append(m)
            meantime = np.array(meantime)

            # Remove bins with zero counts
            allnonzeromask = []
            for bkgsel in self._backgroundselections:
                allnonzeromask.append(np.logical_and(meantime >= bkgsel[0],
                                                     meantime <= bkgsel[1]))

            nonzeromask = allnonzeromask[0]
            if len(allnonzeromask) > 1:
                for mask in allnonzeromask[1:]:
                    nonzeromask = np.logical_or(mask, nonzeromask)

            self.optimalPolGrade = self._fitGlobalAndDetermineOptimumGrade(cnts[nonzeromask], meantime[nonzeromask])

        else:

            self.optimalPolGrade = self._polyorder

        polynomials = []

        for chan in range(self.ttefile.nchans):

            # index all events for this channel and select them
            channelmask = self.ttefile.pha == chan
            # channelselectedevents = self.ttefile.events[channelmask]
            # Mask background events and current channel
            bkgchanmask = np.logical_and(backgroundmask, channelmask)
            # Select the masked events
            currentevents = self.ttefile.events[bkgchanmask]

            binwidth = 1.
            cnts, bins = np.histogram(currentevents - self.ttefile.triggertime,
                                      bins=np.arange(self.ttefile.startevents - self.ttefile.triggertime,
                                                     self.ttefile.stopevents - self.ttefile.triggertime,
                                                     binwidth))

            cnts = cnts / binwidth

            # Find the mean time of the bins
            meantime = []
            for i in xrange(len(bins) - 1):
                m = np.mean((bins[i], bins[i + 1]))
                meantime.append(m)
            meantime = np.array(meantime)

            # Remove bins with zero counts
            allnonzeromask = []
            for bkgsel in self._backgroundselections:
                allnonzeromask.append(np.logical_and(meantime >= bkgsel[0],
                                                     meantime <= bkgsel[1]))

            nonzeromask = allnonzeromask[0]
            if len(allnonzeromask) > 1:
                for mask in allnonzeromask[1:]:
                    nonzeromask = np.logical_or(mask, nonzeromask)

            # Finally, we fit the background and add the polynomial to a list
            thispolynomial, cstat = self._fitChannel(cnts[nonzeromask], meantime[nonzeromask], self.optimalPolGrade)
            polynomials.append(thispolynomial)

        self._polynomials = polynomials

    def _fitChannel(self, cnts, bins, polGrade):

        Nintervals = len(bins)
        # Put data to fit in an x vector and y vector

        polynomial, minLogLike = self._polyfit(bins, cnts, polGrade)

        return polynomial, minLogLike

    def _polyfit(self, x, y, polGrade):

        test = False

        # Check that we have enough counts to perform the fit, otherwise
        # return a "zero polynomial"
        nonzeroMask = (y > 0)
        Nnonzero = len(nonzeroMask.nonzero()[0])
        if (Nnonzero == 0):
            # No data, nothing to do!
            return Polynomial([0.0]), 0.0
        pass

        # Compute an initial guess for the polynomial parameters,
        # with a least-square fit (with weight=1) using SVD (extremely robust):
        # (note that polyfit returns the coefficient starting from the maximum grade,
        # thus we need to reverse the order)
        if (test):
            print("  Initial estimate with SVD..."),
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            initialGuess = np.polyfit(x, y, polGrade)
        pass
        initialGuess = initialGuess[::-1]
        if (test):
            print("  done -> %s" % (initialGuess))

        polynomial = Polynomial(initialGuess)

        # Check that the solution found is meaningful (i.e., definite positive
        # in the interval of interest)
        M = polynomial(x)
        negativeMask = (M < 0)
        if (len(negativeMask.nonzero()[0]) > 0):
            # Least square fit failed to converge to a meaningful solution
            # Reset the initialGuess to reasonable value
            initialGuess[0] = np.mean(y)
            meanx = np.mean(x)
            initialGuess = map(lambda x: abs(x[1]) / pow(meanx, x[0]), enumerate(initialGuess))

        # Improve the solution using a logLikelihood statistic (Cash statistic)
        logLikelihood = BkgLogLikelihood(x, y, polynomial)

        # Check that we have enough non-empty bins to fit this grade of polynomial,
        # otherwise lower the grade
        dof = Nnonzero - (polGrade + 1)
        if (test):
            print("Effective dof: %s" % (dof))
        if (dof <= 2):
            # Fit is poorly or ill-conditioned, have to reduce the number of parameters
            while (dof < 2 and len(initialGuess) > 1):
                initialGuess = initialGuess[:-1]
                polynomial = Polynomial(initialGuess)
                logLikelihood = BkgLogLikelihood(x, y, polynomial)
            pass
        pass

        # Try to improve the fit with the log-likelihood
        # try:
        if (1 == 1):
            finalEstimate = scipy.optimize.fmin(logLikelihood, initialGuess,
                                                ftol=1E-5, xtol=1E-5,
                                                maxiter=1e6, maxfun=1E6,
                                                disp=False)
        # except:
        else:
            # We shouldn't get here!
            raise RuntimeError("Fit failed! Try to reduce the degree of the polynomial.")
        pass

        # Get the value for cstat at the minimum
        minlogLikelihood = logLikelihood(finalEstimate)

        # Update the polynomial with the fitted parameters,
        # and the relative covariance matrix
        finalPolynomial = Polynomial(finalEstimate)
        try:
            finalPolynomial.computeCovarianceMatrix(logLikelihood.getFreeDerivs)
        except Exception:
            raise
        # if test is defined, compare the results with those obtained with ROOT


        return finalPolynomial, minlogLikelihood


class GBMTTEFile(object):
    def __init__(self, ttefile):
        '''
        A simple class for opening and easily accessing Fermi GBM
        TTE Files.

        :param ttefile: The filename of the TTE file to be stored

        '''

        tte = pyfits.open(ttefile)

        self.events = tte['EVENTS'].data['TIME']
        self.pha = tte['EVENTS'].data['PHA']
        self.triggertime = tte['PRIMARY'].header['TRIGTIME']
        self.startevents = tte['PRIMARY'].header['TSTART']
        self.stopevents = tte['PRIMARY'].header['TSTOP']
        self.nchans = tte['EBOUNDS'].header['NAXIS2']

        self._calculate_deattime()


    def _calculate_deattime(self):

        self.deadtime = np.zeros_like(self.events)
        overflowmask =  self.pha == 128

        # Dead time for overflow (note, overflow sometimes changes)
        self.deadtime[overflowmask] = 10.E-6 #s

        # Normal dead time
        self.deadtime[~overflowmask] = 2.E-6 #s


class BkgLogLikelihood(object):
    '''
    Implements a Poisson likelihood (i.e., the Cash statistic). Mind that this is not
    the Castor statistic (Cstat). The difference between the two is a constant given
    a dataset. I kept Cash instead of Castor to make easier the comparison with ROOT
    during tests, since ROOT implements the Cash statistic.
    '''

    def __init__(self, x, y, model, **kwargs):
        self.x = x
        self.y = y
        self.model = model
        self.parameters = model.getParams()

        # Initialize the exposure to 1.0 (i.e., non-influential)
        # It will be replaced by the real exposure if the exposure keyword
        # have been used
        self.exposure = np.zeros(len(x)) + 1.0

        for key in kwargs.keys():
            if (key.lower() == "exposure"):
                self.exposure = np.array(kwargs[key])
        pass

    pass

    def _evalLogM(self, M):
        # Evaluate the logarithm with protection for negative or small
        # numbers, using a smooth linear extrapolation (better than just a sharp
        # cutoff)
        tiny = np.float64(np.finfo(M[0]).tiny)

        nontinyMask = (M > 2.0 * tiny)
        tinyMask = np.logical_not(nontinyMask)

        if (len(tinyMask.nonzero()[0]) > 0):
            logM = np.zeros(len(M))
            logM[tinyMask] = np.abs(M[tinyMask]) / tiny + np.log(tiny) - 1
            logM[nontinyMask] = np.log(M[nontinyMask])
        else:
            logM = np.log(M)
        return logM

    pass

    def __call__(self, parameters):
        '''
          Evaluate the Cash statistic for the given set of parameters
        '''

        # Compute the values for the model given this set of parameters
        self.model.setParams(parameters)
        M = self.model(self.x) * self.exposure
        Mfixed, tiny = self._fixPrecision(M)

        # Replace negative values for the model (impossible in the Poisson context)
        # with zero
        negativeMask = (M < 0)
        if (len(negativeMask.nonzero()[0]) > 0):
            M[negativeMask] = 0.0
        pass

        # Poisson loglikelihood statistic (Cash) is:
        # L = Sum ( M_i - D_i * log(M_i))

        logM = self._evalLogM(M)

        # Evaluate v_i = D_i * log(M_i): if D_i = 0 then the product is zero
        # whatever value has log(M_i). Thus, initialize the whole vector v = {v_i}
        # to zero, then overwrite the elements corresponding to D_i > 0
        d_times_logM = np.zeros(len(self.y))
        nonzeroMask = (self.y > 0)
        d_times_logM[nonzeroMask] = self.y[nonzeroMask] * logM[nonzeroMask]

        logLikelihood = np.sum(Mfixed - d_times_logM)

        return logLikelihood

    pass

    def _fixPrecision(self, v):
        '''
          Round extremely small number inside v to the smallest usable
          number of the type corresponding to v. This is to avoid warnings
          and errors like underflows or overflows in math operations.
        '''
        tiny = np.float64(np.finfo(v[0]).tiny)
        zeroMask = (np.abs(v) <= tiny)
        if (len(zeroMask.nonzero()[0]) > 0):
            v[zeroMask] = np.sign(v[zeroMask]) * tiny

        return v, tiny

    pass

    def getFreeDerivs(self, parameters=None):
        '''
        Return the gradient of the logLikelihood for a given set of parameters (or the current
        defined one, if parameters=None)
        '''
        # The derivative of the logLikelihood statistic respect to parameter p is:
        # dC / dp = Sum [ (dM/dp)_i - D_i/M_i (dM/dp)_i]

        # Get the number of parameters and initialize the gradient to 0
        Nfree = self.model.getNumFreeParams()
        derivs = np.zeros(Nfree)

        # Set the parameters, if a new set has been provided
        if (parameters != None):
            self.model.setParams(parameters)
        pass

        # Get the gradient of the model respect to the parameters
        modelDerivs = self.model.getFreeDerivs(self.x) * self.exposure
        # Get the model
        M = self.model(self.x) * self.exposure

        M, tinyM = self._fixPrecision(M)

        # Compute y_divided_M = y/M: inizialize y_divided_M to zero
        # and then overwrite the elements for which y > 0. This is to avoid
        # possible underflow and overflow due to the finite precision of the
        # computer
        y_divided_M = np.zeros(len(self.y))
        nonzero = (self.y > 0)
        y_divided_M[nonzero] = self.y[nonzero] / M[nonzero]

        for p in range(Nfree):
            thisModelDerivs, tinyMd = self._fixPrecision(modelDerivs[p])
            derivs[p] = np.sum(thisModelDerivs * (1.0 - y_divided_M))
        pass

        return derivs

    pass


pass


class Polynomial(object):
    def __init__(self, params):
        self.params = params
        self.degree = len(params) - 1

        # Build an empty covariance matrix
        self.covMatrix = np.zeros([self.degree + 1, self.degree + 1])

    pass

    def horner(self, x):
        """A function that implements the Horner Scheme for evaluating a
        polynomial of coefficients *args in x."""
        result = 0
        for coefficient in self.params[::-1]:
            result = result * x + coefficient
        return result

    pass

    def __call__(self, x):
        return self.horner(x)

    pass

    def __str__(self):
        # This is call by the print() command
        # Print results
        output = "\n------------------------------------------------------------"
        output += '\n| {0:^10} | {1:^20} | {2:^20} |'.format("COEFF", "VALUE", "ERROR")
        output += "\n|-----------------------------------------------------------"
        for i, parValue in enumerate(self.getParams()):
            output += '\n| {0:<10d} | {1:20.5g} | {2:20.5g} |'.format(i, parValue, math.sqrt(self.covMatrix[i, i]))
        pass
        output += "\n------------------------------------------------------------"

        return output

    pass

    def setParams(self, parameters):
        self.params = parameters

    pass

    def getParams(self):
        return self.params

    pass

    def getNumFreeParams(self):
        return self.degree + 1

    pass

    def getFreeDerivs(self, x):
        Npar = self.degree + 1
        freeDerivs = []
        for i in range(Npar):
            freeDerivs.append(map(lambda xx: pow(xx, i), x))
        pass
        return np.array(freeDerivs)

    pass

    def computeCovarianceMatrix(self, statisticGradient):
        self.covMatrix = computeCovarianceMatrix(statisticGradient, self.params)
        # Check that the covariance matrix is positive-defined
        negativeElements = (np.matrix.diagonal(self.covMatrix) < 0)
        if (len(negativeElements.nonzero()[0]) > 0):
            raise RuntimeError(
                "Negative element in the diagonal of the covariance matrix. Try to reduce the polynomial grade.")

    pass

    def getCovarianceMatrix(self):
        return self.covMatrix

    pass

    def integral(self, xmin, xmax):
        '''
        Evaluate the integral of the polynomial between xmin and xmax
        '''
        integralCoeff = [0]
        integralCoeff.extend(map(lambda i: self.params[i - 1] / float(i), range(1, self.degree + 1 + 1)))

        integralPolynomial = Polynomial(integralCoeff)

        return integralPolynomial(xmax) - integralPolynomial(xmin)

    pass

    def integralError(self, xmin, xmax):
        # Based on http://root.cern.ch/root/html/tutorials/fit/ErrorIntegral.C.html

        # Set the weights
        i_plus_1 = np.array(range(1, self.degree + 1 + 1), 'd')

        def evalBasis(x):
            return (1 / i_plus_1) * pow(x, i_plus_1)

        c = evalBasis(xmax) - evalBasis(xmin)

        # Compute the error on the integral
        err2 = 0.0
        nPar = self.degree + 1
        parCov = self.getCovarianceMatrix()
        for i in range(nPar):
            s = 0.0
            for j in range(nPar):
                s += parCov[i, j] * c[j]
            pass
            err2 += c[i] * s
        pass

        return math.sqrt(err2)

    pass


pass


def computeCovarianceMatrix(grad, par, full_output=False,
                            init_step=0.01, min_step=1e-12, max_step=1, max_iters=50,
                            target=0.1, min_func=1e-7, max_func=4):
    """Perform finite differences on the _analytic_ gradient provided by user to calculate hessian/covariance matrix.

    Positional args:
        grad                : a function to return a gradient
        par                 : vector of parameters (should be function minimum for covariance matrix calculation)

    Keyword args:

        full_output [False] : if True, return information about convergence, else just the covariance matrix
        init_step   [1e-3]  : initial step size (0.04 ~ 10% in log10 space); can be a scalar or vector
        min_step    [1e-6]  : the minimum step size to take in parameter space
        max_step    [1]     : the maximum step size to take in parameter sapce
        max_iters   [5]     : maximum number of iterations to attempt to converge on a good step size
        target      [0.5]   : the target change in the function value for step size
        min_func    [1e-4]  : the minimum allowable change in (abs) function value to accept for convergence
        max_func    [4]     : the maximum allowable change in (abs) function value to accept for convergence
    """

    nparams = len(par)
    step_size = np.ones(nparams) * init_step
    step_size = np.maximum(step_size, min_step * 1.1)
    step_size = np.minimum(step_size, max_step * 0.9)
    hess = np.zeros([nparams, nparams])
    min_flags = np.asarray([False] * nparams)
    max_flags = np.asarray([False] * nparams)

    def revised_step(delta_f, current_step, index):
        if (current_step == max_step):
            max_flags[i] = True
            return True, 0

        elif (current_step == min_step):
            min_flags[i] = True
            return True, 0

        else:
            adf = abs(delta_f)
            if adf < 1e-8:
                # need to address a step size that results in a likelihood change that's too
                # small compared to precision
                pass

            if (adf < min_func) or (adf > max_func):
                new_step = current_step / (adf / target)
                new_step = min(new_step, max_step)
                new_step = max(new_step, min_step)
                return False, new_step
            else:
                return True, 0

    iters = np.zeros(nparams)
    for i in xrange(nparams):
        converged = False

        for j in xrange(max_iters):
            iters[i] += 1

            di = step_size[i]
            par[i] += di
            g_up = grad(par)

            par[i] -= 2 * di
            g_dn = grad(par)

            par[i] += di

            delta_f = (g_up - g_dn)[i]

            converged, new_step = revised_step(delta_f, di, i)
            # print 'Parameter %d -- Iteration %d -- Step size: %.2e -- delta: %.2e'%(i,j,di,delta_f)

            if converged:
                break
            else:
                step_size[i] = new_step
        pass

        hess[i, :] = (g_up - g_dn) / (2 * di)  # central difference

        if not converged:
            print 'Warning: step size for parameter %d (%.2g) did not result in convergence.' % (i, di)
    try:
        cov = np.linalg.inv(hess)
    except:
        print 'Error inverting hessian.'
        raise Exception('Error inverting hessian')
    if full_output:
        return cov, hess, step_size, iters, min_flags, max_flags
    else:
        return cov
    pass


pass



import numdifftools as nd
import numpy as np

from astromodels import SettingOutOfBounds


class ParameterOnBoundary(RuntimeError):
    pass


class CannotComputeHessian(RuntimeError):
    pass


def get_hessian(function, point, minima, maxima):

    # Define a wrapper because numdifftools expect the function to be f(x) with
    # x a vector, while the likelihood function expect f(x1,x2,x3...)

    # Also, compute the orders of magnitude of the parameters, which we will use to make the values used
    # by numdifftools close to 1

    try:

        n_dim = point.shape[0]
        _ = minima.shape
        _ = maxima.shape

    except AttributeError:

        point = np.array(point, ndmin=1)
        minima = np.array(minima, ndmin=1)
        maxima = np.array(maxima, ndmin=1)

        n_dim = point.shape[0]

    orders_of_magnitude = 10**np.ceil(np.log10(np.abs(point)))

    scaled_point = point / orders_of_magnitude
    scaled_minima = minima / orders_of_magnitude
    scaled_maxima = maxima / orders_of_magnitude

    def wrapper(x):

        scaled_back_x = x * orders_of_magnitude

        try:

            result = function(*scaled_back_x)

        except SettingOutOfBounds:

            raise CannotComputeHessian("Cannot compute Hessian, parameters out of bounds at %s" % scaled_back_x)

        else:

            return result

    # Decide a delta for the finite differentiation
    # The algorithm implemented in numdifftools is robust with respect to the choice
    # of delta, as long as we are not going beyond the boundaries (which would cause
    # the procedure to fail)

    scaled_deltas = np.zeros_like(scaled_point)

    for i in range(n_dim):

        scaled_value = scaled_point[i]

        scaled_min_value, scaled_max_value = (scaled_minima[i], scaled_maxima[i])

        if scaled_value == scaled_min_value or scaled_value == scaled_max_value:

            raise ParameterOnBoundary("Value for parameter number %s is on the boundary" % i)

        if scaled_min_value is not None:

            # Parameter with low bound

            distance_to_min = scaled_value - scaled_min_value

        else:

            # No defined minimum

            distance_to_min = np.inf

        if scaled_max_value is not None:

            # Parameter with hi bound

            distance_to_max = scaled_max_value - scaled_value

        else:

            # No defined maximum

            distance_to_max = np.inf

        # Delta is the minimum between 3% of the value, and half of the minimum
        # distance to either boundary

        scaled_deltas[i] = min([0.03 * abs(scaled_point[i]), distance_to_max / 2.0, distance_to_min / 2.0])

    # Compute the Hessian matrix at best_fit_values

    hessian_matrix_ = nd.Hessian(wrapper, scaled_deltas)(scaled_point)

    # Transform it to numpy matrix

    hessian_matrix = np.array(hessian_matrix_)

    # Now correct back the Hessian for the scales
    for i in range(n_dim):

        for j in range(n_dim):

            hessian_matrix[i,j] /= orders_of_magnitude[i] * orders_of_magnitude[j]

    return hessian_matrix


#Author: Giacomo Vianello (giacomov@slac.stanford.edu)

#A hack on the astropy Table class to make its output
#more appealing, especially when in the Ipython notebook

import astropy.table

class Table(astropy.table.Table):
    
    def _base_repr_(self, html=False, show_name=True, **kwargs):
        '''Override the method in the astropy.Table class
        to avoid displaying the description, and the format
        of the columns'''
        
        tableid = 'table{id}'.format(id=id(self))
        
        data_lines, outs = self.formatter._pformat_table(self,
            tableid=tableid, html=html, max_width=(-1 if html else None),
            show_name=show_name, show_unit=None, show_dtype=False)
                
        out = '\n'.join(data_lines)
        if astropy.table.six.PY2 and isinstance(out, astropy.table.six.text_type):
            out = out.encode('utf-8')

        return out


class NumericMatrix(Table):
    
    def _base_repr_(self, html=False,**kwargs):
        return super(NumericMatrix, self)._base_repr_(html, show_name=False, **kwargs)
    



import UnbinnedAnalysis
import BinnedAnalysis
import pyLikelihood as pyLike
import numpy
from GtBurst import LikelihoodComponent
from GtBurst import FuncFactory
import matplotlib.pyplot as plt
from matplotlib import gridspec

from threeML.plugin_prototype import PluginPrototype
from astromodels import Parameter
from threeML.plugins.gammaln import logfactorial

__instrument_name = "Fermi LAT (standard classes)"


class myPointSource(LikelihoodComponent.GenericSource):
    def __init__(self, source, name):
        self.source = source
        self.source.name = name


class LikelihoodModelConverter(object):
    def __init__(self, likelihoodModel, irfs):

        self.likelihoodModel = likelihoodModel

        self.irfs = irfs

    def setFileSpectrumEnergies(self, emin_kev, emax_kev, nEnergies):

        self.energiesKeV = numpy.logspace(numpy.log10(emin_kev),
                                          numpy.log10(emax_kev),
                                          nEnergies)

    def writeXml(self, xmlfile, ra, dec, roi):

        # Loop through all the sources in the likelihood model and generate a FileSpectrum
        # for all of them. This is necessary to allow the FermiLATLike class
        # to update the spectrum in pyLikelihood without having to write and read a .xml file
        # on the disk

        allSourcesForPyLike = []

        nPtsrc = self.likelihoodModel.get_number_of_point_sources()

        for ip in range(nPtsrc):
            allSourcesForPyLike.append(self._makeFileSpectrum(ip))

        # Now the same for extended sources

        nExtSrc = self.likelihoodModel.get_number_of_extended_sources()

        if (nExtSrc > 0):
            raise NotImplemented("Cannot support extended sources yet!")

        iso = LikelihoodComponent.IsotropicTemplate(self.irfs)

        iso.source.spectrum.Normalization.max = 1.5
        iso.source.spectrum.Normalization.min = 0.5
        iso.source.spectrum.setAttributes()

        allSourcesForPyLike.append(iso)

        gal = LikelihoodComponent.GalaxyAndExtragalacticDiffuse(
            self.irfs, ra, dec, 2.5 * roi)
        gal.source.spectrum.Value.max = 1.5
        gal.source.spectrum.Value.min = 0.5
        gal.source.spectrum.setAttributes()

        allSourcesForPyLike.append(gal)

        # Now generate the xml file with also the Galactic and Isotropic diffuse
        # templates
        xml = LikelihoodComponent.LikelihoodModel()
        xml.addSources(*allSourcesForPyLike)
        xml.writeXML(xmlfile)

    def _makeFileSpectrum(self, ip):

        name = self.likelihoodModel.get_point_source_name(ip)

        values = self.likelihoodModel.get_point_source_fluxes(ip,
                                                              self.energiesKeV)

        tempName = "__%s.txt" % name

        with open(tempName, "w+") as f:
            for e, v in zip(self.energiesKeV, values):
                # Gtlike needs energies in MeV and fluxes in ph/MeV/cm2)

                f.write("%s %s\n" % (e / 1000.0, v * 1000.0))

        # p                         = fileFunction.parameter("Normalization")
        # p.setBounds(1-float(effAreaAllowedSize),1+effAreaAllowedSize)

        # Now generate the XML source wrapper
        # This is convoluted, but that's the ST way of doing things...
        # The final product is a class with a writeXml method

        src = '\n'.join((('<source name= "%s" ' % name) + 'type="PointSource">',
                         '   <spectrum type="PowerLaw2"/>',
                         '   <!-- point source units are cm^-2 s^-1 MeV^-1 -->',
                         '   <spatialModel type="SkyDirFunction"/>',
                         '</source>\n'))
        src = FuncFactory.minidom.parseString(src).getElementsByTagName('source')[0]
        src = FuncFactory.Source(src)

        src.spectrum = FuncFactory.FileFunction()
        src.spectrum.file = tempName
        src.spectrum.parameters['Normalization'].value = 1.0
        src.spectrum.parameters['Normalization'].max = 1.1
        src.spectrum.parameters['Normalization'].min = 0.9
        src.spectrum.setAttributes()
        src.deleteChildElements('spectrum')
        src.node.appendChild(src.spectrum.node)

        src.spatialModel = FuncFactory.SkyDirFunction()
        src.deleteChildElements('spatialModel')
        src.node.appendChild(src.spatialModel.node)

        ra, dec = self.likelihoodModel.get_point_source_position(ip)

        src.spatialModel.RA.value = ra
        src.spatialModel.DEC.value = dec
        src.spatialModel.setAttributes()
        src.setAttributes()

        return myPointSource(src, name)


class FermiLATLike(PluginPrototype):
    def __init__(self, name, eventFile, ft2File, livetimeCube, kind, *args):

        self.name = name

        # Read the ROI cut
        cc = pyLike.RoiCuts()
        cc.readCuts(eventFile, "EVENTS")
        self.ra, self.dec, self.rad = cc.roiCone()

        # Read the IRF selection
        c = pyLike.Cuts(eventFile, "EVENTS")
        self.irf = c.CALDB_implied_irfs()

        self.ft2File = ft2File
        self.livetimeCube = livetimeCube

        # These are the boundaries and the number of energies for the computation
        # of the model (in keV)
        self.emin = 1e4
        self.emax = 5e8
        self.Nenergies = 1200

        # This is the limit on the effective area correction factor,
        # which is a multiplicative factor in front of the whole model
        # to account for inter-calibration issues. By default it can vary
        # by 10%. This can be changed by issuing:
        # FermiLATUnbinnedLikeInstance.effCorrLimit = [new limit]
        # where for example a [new limit] of 0.2 allow for an effective
        # area correction up to +/- 20 %

        self.effCorrLimit = 0.1

        if (kind.upper() != "UNBINNED" and kind.upper() != "BINNED"):

            raise RuntimeError("Accepted values for the kind parameter are: " +
                               "binned, unbinned. You specified: %s" % (kind))

        else:

            self.kind = kind.upper()

        if (kind.upper() == 'UNBINNED'):
            exposureMap = args[0]
            self.eventFile = eventFile
            self.exposureMap = exposureMap
            # Read the files and generate the pyLikelihood object
            self.obs = UnbinnedAnalysis.UnbinnedObs(self.eventFile,
                                                    self.ft2File,
                                                    expMap=self.exposureMap,
                                                    expCube=self.livetimeCube,
                                                    irfs=self.irf)

        elif (kind.upper() == "BINNED"):
            sourceMaps, binnedExpoMap = args
            self.sourceMaps = sourceMaps
            self.binnedExpoMap = binnedExpoMap

            self.obs = BinnedAnalysis.BinnedObs(srcMaps=self.sourceMaps,
                                                expCube=self.livetimeCube,
                                                binnedExpMap=self.binnedExpoMap,
                                                irfs=self.irf)
        pass

        # Activate inner minimization by default
        self.setInnerMinimization(True)

    pass

    def set_model(self, likelihoodModel):
        '''
        Set the model to be used in the joint minimization.
        Must be a LikelihoodModel instance.
        '''
        self.lmc = LikelihoodModelConverter(likelihoodModel,
                                            self.irf)

        self.lmc.setFileSpectrumEnergies(self.emin, self.emax, self.Nenergies)

        xmlFile = '__xmlFile.xml'

        self.lmc.writeXml(xmlFile, self.ra, self.dec, self.rad)

        if (self.kind == "BINNED"):
            self.like = BinnedAnalysis.BinnedAnalysis(self.obs,
                                                      xmlFile,
                                                      optimizer='DRMNFB')

        else:

            self.like = UnbinnedAnalysis.UnbinnedAnalysis(self.obs,
                                                          xmlFile,
                                                          optimizer='DRMNFB')

        self.likelihoodModel = likelihoodModel

        # Here we need also to compute the logLike value, so that the model
        # in the XML file will be chanded if needed
        dumb = self.get_log_like()

        # Build the list of the nuisance parameters
        self._setNuisanceParameters()

    pass

    def get_name(self):
        '''
        Return a name for this dataset (likely set during the constructor)
        '''
        return self.name

    pass

    def setInnerMinimization(self, s):

        self.innerMinimization = bool(s)

    def inner_fit(self):
        '''
        This is used for the profile likelihood. Keeping fixed all parameters in the
        modelManager, this method minimize the logLike over the remaining nuisance
        parameters, i.e., the parameters belonging only to the model for this
        particular detector
        '''
        self._updateGtlikeModel()

        if not self.innerMinimization:

            log_like = self.like.logLike.value()

        else:

            try:
                # Use .optimize instead of .fit because we don't need the errors
                # (.optimize is faster than .fit)
                self.like.optimize(0)
            except:
                # This is necessary because sometimes fitting algorithms go and explore extreme region of the
                # parameter space, which might turn out to give strange model shapes and therefore
                # problems in the likelihood fit
                print("Warning: failed likelihood fit (probably parameters are too extreme).")
                return 1e5
            else:
                # Update the value for the nuisance parameters
                for par in self.nuisanceParameters.values():
                    newValue = self.getNuisanceParameterValue(par.name)
                    par.value = newValue
                pass

                log_like = self.like.logLike.value()

        return log_like - logfactorial(self.like.total_nobs())

    pass

    # def _initFileFunction(self):
    # If the point source is already in the model, delete it
    # (like.deleteSource() returns the model)
    # gtlikeSrcModel            = self.like.deleteSource(self.modelManager.name)

    # Get the new model for the source (with the latest parameter values)
    # and add it back to the likelihood model
    #  self._getNewGtlikeModel(gtlikeSrcModel,self.effCorrLimit)
    #  self.like.addSource(gtlikeSrcModel)

    # Slow! But no other options at the moment
    #  self.like.writeXml(self.xmlModel)
    #  self.like.logLike.reReadXml(self.xmlModel)
    # pass

    # def _getNewGtlikeModel(self,currentGtlikeModel,effAreaAllowedSize=0.1):
    # Write on disk the current model
    #  tempName                  = os.path.join(
    #                              os.path.dirname(
    #                              os.path.abspath(self.xmlModel)
    #                                             ),
    #                             '__fileSpectrum.txt')
    #
    #  #This will recompute the model if necessary
    #  self.modelManager.writeToFile(tempName,self.emin,self.emax,self.Nenergies)
    #
    #  #Generate a new FileFunction spectrum and assign it to the source
    #  fileFunction              = pyLike.FileFunction()
    #  fileFunction.readFunction(tempName)
    #  fileFunction.setParam("Normalization",1)
    #  p                         = fileFunction.parameter("Normalization")
    #  p.setBounds(1-float(effAreaAllowedSize),1+effAreaAllowedSize)
    #
    #  currentGtlikeModel.setSpectrum(fileFunction)
    # pass

    def _updateGtlikeModel(self):
        '''
        #Slow! But no other options at the moment
        self.like.writeXml(self.xmlModel)
        self.like.logLike.reReadXml(self.xmlModel)
        '''

        energies = self.lmc.energiesKeV

        for id, srcName in enumerate(self.likelihoodModel.point_sources.keys()):

            values = self.likelihoodModel.get_point_source_fluxes(id, energies)

            gtlikeSrcModel = self.like[srcName]

            my_function = gtlikeSrcModel.getSrcFuncs()['Spectrum']
            my_file_function = pyLike.FileFunction_cast(my_function)

            my_file_function.setParam("Normalization", 1)

            # Cap the values to avoid numerical errors

            capped_values = numpy.minimum(numpy.maximum(values * 1000, 1e-25), 1e5)

            my_file_function.setSpectrum(energies / 1000.0, capped_values)
            gtlikeSrcModel.setSpectrum(my_file_function)

            # TODO: extended sources

        self.like.syncSrcParams()

    def get_log_like(self):
        '''
        Return the value of the log-likelihood with the current values for the
        parameters stored in the ModelManager instance
        '''
        self._updateGtlikeModel()
        try:

            value = self.like.logLike.value()

        except:

            raise

        return value - logfactorial(self.like.total_nobs())

    pass

    def getModelAndData(self):

        fake = numpy.array([])

        return fake, fake, fake, fake

    pass

    def display(self):

        e1 = self.like.energies[:-1]
        e2 = self.like.energies[1:]

        ec = (e1 + e2) / 2.0
        de = (e2 - e1) / 2.0

        sum_model = numpy.zeros_like(self.like._srcCnts(self.like.sourceNames()[0]))

        fig = plt.figure()

        gs = gridspec.GridSpec(2, 1, height_ratios=[2, 1])
        gs.update(hspace=0)

        sub = plt.subplot(gs[0])

        for sourceName in self.like.sourceNames():
            sum_model = sum_model + self.like._srcCnts(sourceName)

            sub.plot(ec, self.like._srcCnts(sourceName), label=sourceName)

        sub.plot(ec, sum_model, label='Total Model')

        sub.errorbar(ec, self.like.nobs, xerr=de,
                     yerr=numpy.sqrt(self.like.nobs),
                     fmt='.', label='Counts')

        plt.legend(bbox_to_anchor=(1.05, 1), loc=2, numpoints=1)

        # Residuals

        sub1 = plt.subplot(gs[1])

        # Using model variance to account for low statistic

        resid = (self.like.nobs - sum_model) / sum_model

        sub1.axhline(0, linestyle='--')
        sub1.errorbar(ec, resid, xerr=de,
                      yerr=numpy.sqrt(self.like.nobs) / sum_model,
                      capsize=0, fmt='.')

        sub.set_xscale("log")
        sub.set_yscale("log", nonposy='clip')

        sub.set_ylabel("Counts per bin")

        sub1.set_xscale("log")

        sub1.set_xlabel("Energy (MeV)")
        sub1.set_ylabel("(Mo. - data) / mo.")

        sub.set_xticks([])

        return fig

    def _setNuisanceParameters(self):

        # Get the list of the sources
        sources = list(self.like.model.srcNames)

        freeParamNames = []
        for srcName in sources:
            thisNamesV = pyLike.StringVector()
            thisSrc = self.like.logLike.getSource(srcName)
            thisSrc.spectrum().getFreeParamNames(thisNamesV)
            thisNames = map(lambda x: "%s-%s" % (srcName, x), thisNamesV)
            freeParamNames.extend(thisNames)
        pass

        self.nuisanceParameters = {}

        for name in freeParamNames:
            value = self.getNuisanceParameterValue(name)
            bounds = self.getNuisanceParameterBounds(name)
            delta = self.getNuisanceParameterDelta(name)
            self.nuisanceParameters[name] = Parameter(name, value, min_value=bounds[0],
                                                      max_value=bounds[1], delta=delta)
        pass

    pass

    def get_nuisance_parameters(self):
        '''
        Return a list of nuisance parameters. Return an empty list if there
        are no nuisance parameters
        '''
        return self.nuisanceParameters

    def getNuisanceParameterValue(self, paramName):
        src, pname = paramName.split("-")
        return self.like.model[src].funcs['Spectrum'].getParam(pname).getValue()

    pass

    def getNuisanceParameterBounds(self, paramName):
        src, pname = paramName.split("-")
        return list(self.like.model[src].funcs['Spectrum'].getParam(pname).getBounds())

    pass

    def getNuisanceParameterDelta(self, paramName):
        src, pname = paramName.split("-")
        value = self.like.model[src].funcs['Spectrum'].getParam(pname).getValue()
        return value / 100.0

    pass

    def setNuisanceParameterValue(self, paramName, value):
        src, pname = paramName.split("-")
        self.like.model[src].funcs['Spectrum'].getParam(pname).setValue(value)

    pass


pass



import warnings

import numpy as np

from threeML.plugin_prototype import PluginPrototype
from threeML.plugins.ogip import OGIPPHA, OGIPPluginCash

__instrument_name = "All OGIP-compliant instruments"


class GenericOGIPLike(OGIPPluginCash, PluginPrototype):

    def __init__(self, name, phafile, bkgfile, rspfile, arffile=None):
        """
        If the input files are PHA2 files, remember to specify the spectrum number, for example:
        FermiGBMLike("GBM","spectrum.pha{2}","bkgfile.bkg{2}","rspfile.rsp{2}")
        to load the second spectrum, second background spectrum and second response.
        """

        OGIPPluginCash.__init__(self, name, rspfile, arffile=arffile, bkgfile=bkgfile, phafile=phafile)

        self.phafile = OGIPPHA(phafile, filetype='observed')
        exposure = self.phafile.getExposure()
        self.bkgfile = OGIPPHA(bkgfile, filetype="background")

        # Start with an empty mask (the user will overwrite it using the
        # setActiveMeasurement method)
        mask = np.asarray(np.ones(self.phafile.getRates().shape), np.bool)

        # Get the counts for this spectrum
        counts = (self.phafile.getRates() * exposure)

        # Get the background counts for this spectrum
        bkgCounts = (self.bkgfile.getRates() * exposure)

        self._initialSetup(mask, counts, bkgCounts, exposure)


import numpy
from scipy.special import gammaln

def _gammaln(x):
    """Vectorized calculation of ln(abs(gamma(array))) across a Numpy array.

    Numpy does not have a native implementation of gammaln.
    U{Scipy does <http://docs.scipy.org/doc/scipy/reference/generated/scipy.special.gammaln.html>},
    but that would introduce a dependency.
    """
    
    array       = numpy.asarray(x)
    gammaln_cof = [76.18009173, -86.50532033, 24.01409822, -1.231739516e0, 0.120858003e-2, -0.536382e-5]
    gammaln_stp = 2.50662827465
    x = numpy.array(array - 1.0)
    tmp = x + 5.5
    tmp = ((x + 0.5)*numpy.log(tmp)) - tmp
    ser = numpy.ones(array.shape[0], dtype=numpy.dtype(float))
    for cof in gammaln_cof:
        x += 1.0
        ser += cof/x
    return (tmp + numpy.log(gammaln_stp*ser))

def logfactorial(n):
    
    return gammaln( n + 1)



import numpy

class Binner( object ):
    
    def __init__( self, edges1, edges2, counts ):
        
        self.counts = numpy.array( counts )
        
        self.e1 = numpy.array( edges1 )
        self.e2 = numpy.array( edges2 )
        
        self.total = numpy.sum( self.counts )
    
    def byConstantCounts( self, n, secondVector = None ):
        '''
        Bin data to have an equal number of counts in each bin
        '''
        
        #Integral distribution
        
        intDistr = numpy.cumsum( self.counts )
        
        if secondVector is not None:
            
            intDistr2 = numpy.cumsum( secondVector )
        
        #Get the indexes of the elements in intDistr
        #where the value changes by more than n
        
        idx = []
        values = []
        newV = []
        
        end = []
        
        reference = 0
        
        ref2 = 0
        
        for i, d in enumerate( intDistr ):
            
            value = d - reference
            
            if value >= n:
                
                idx.append( i )
                
                values.append( value )
                
                reference = d
                
                end.append( self.e2[i] )
                
                if secondVector is not None:
                    
                    newV.append( intDistr2[i] - ref2 )
                    
                    ref2 = intDistr2[i]
                
                
        #Always add the last point
        if self.counts.shape[0] - 1 not in idx:
            
            idx.append( self.counts.shape[0] - 1)
            
            values.append( self.total - reference )
            
            if secondVector is not None:
                
                newV.append( numpy.sum(secondVector) - ref2 )
            
            end.append( self.e2[-1] )
        
        #Now create the start of the edges
        
        start = [ self.e1[0] ]
        start.extend( end[:-1] )
        
        #Safety check
        
        if numpy.sum( values ) != self.total:
            
            raise RuntimeError("This is a bug in the Binner. The total counts are not the same as the input")
        
        return numpy.array( start ), numpy.array( end ), numpy.array( values ), numpy.array( newV )



from contextlib import contextmanager

has_parallel = False

try:

    from ipyparallel import Client

except ImportError:

    # Try with the old ipython, which contained the parallel code
    try:

        from IPython.parallel import Client

    except ImportError:

        has_parallel = False

    else:

        has_parallel = True

else:

    has_parallel = True

from threeML.config.config import threeML_config

# Custom warning
import warnings
import exceptions

class NoParallelEnvironment(exceptions.UserWarning):
    pass

# Set up the warnings module to always display our custom warning (otherwise it would only be displayed once)
warnings.simplefilter('always', NoParallelEnvironment)

@contextmanager
def parallel_computation(profile=None):
    """
    A context manager which turns on parallel execution temporarily

    :param profile: the profile to use, if different from the default
    :return:
    """

    # Memorize the state of the use-parallel config

    old_state = bool(threeML_config['parallel']['use-parallel'])

    old_profile = str(threeML_config['parallel']['IPython profile name'])

    # Set the use-parallel feature on, if available

    if has_parallel:

        threeML_config['parallel']['use-parallel'] = True

    else:

        # No parallel environment available. Issue a warning and continue with serial computation

        warnings.warn("You requested parallel computation, but no parallel environment is available. You need "
                      "to install the ipyparallel package. Continuing with serial computation...",
                      NoParallelEnvironment)

        threeML_config['parallel']['use-parallel'] = False

    # Now use the specified profile (if any), otherwise the default one

    if profile is not None:

        threeML_config['parallel']['IPython profile name'] = str(profile)

    # Here is where the content of the with parallel_computation statement gets
    # executed

    try:

        yield

    finally:

        # This gets executed in any case, even if there is an exception

        # Revert back
        threeML_config['parallel']['use-parallel'] = old_state

        threeML_config['parallel']['IPython profile name'] = old_profile


if has_parallel:
    class ParallelClient(Client):

        def __init__(self, *args, **kwargs):
            """
            Wrapper around the IPython Client class, which forces the use of dill for object serialization

            :param args: same as IPython Client
            :param kwargs: same as IPython Client
            :return:
            """

            import pdb;pdb.set_trace()

            # Just a wrapper around the IPython Client class
            # forcing the use of dill for object serialization
            # (more robust, and allows for serialization of class
            # methods)

            if 'profile' not in kwargs.keys():

                kwargs['profile'] = threeML_config['parallel']['IPython profile name']

            super(ParallelClient, self).__init__(*args, **kwargs)

            # This will propagate the use_dill to all running
            # engines
            _ = self.direct_view().use_dill()

        def get_number_of_engines(self):

            return len(self.direct_view())

else:

    # NO parallel environment available. Make a dumb object to avoid import problems, but this object will never
    # be really used because the context manager will not activate the parallel mode (see above)
    class ParallelClient(object):

        def __init__(self, *args, **kwargs):

            raise RuntimeError("No parallel environment and attempted to use the ParallelClient class, which should "
                               "never happen. Please open an issue at https://github.com/giacomov/3ML/issues")
